SystemOrganization addCategory: #'Islands-Kernel'!
SystemOrganization addCategory: #'Islands-Support'!
SystemOrganization addCategory: #'Islands-Files'!
SystemOrganization addCategory: #'Islands-Time'!

!BalloonBuffer class methodsFor: '*Islands' stamp: 'ar 1/16/2005 21:54'!
howToPassAsArgument
	"passByIdentity: for now - these buffers are shared for rendering"
	^#passByIdentity:! !

!MorphicEventDispatcher class methodsFor: '*Islands' stamp: 'ar 1/16/2005 17:32'!
howToPassAsArgument
	^#passByCopy:! !

!IdentityTransform class methodsFor: '*Islands' stamp: 'ar 1/16/2005 22:17'!
howToPassAsArgument
	"IdentityTransforms are immutable"
	^#passByIdentity:! !

!Boolean class methodsFor: '*Islands' stamp: 'ar 1/16/2005 21:55'!
howToPassAsArgument
	"passByIdentity since it is immutable"
	^#passByIdentity:! !

!Boolean methodsFor: '*Islands' stamp: 'ar 6/16/2005 19:39'!
exportOnIslandWriter: writer
	^writer exportBoolean: self! !

!Delay class methodsFor: '*Islands' stamp: 'ar 1/16/2005 21:57'!
howToPassAsArgument
	"passByIdentity: for now; I think we don't need this but some primitives are unhappy with far refs and need some checking before this will work properly"
	^#passByIdentity:! !

TestCase subclass: #IslandTests
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Islands-Support'!

!IslandTests commentStamp: 'ar 1/17/2005 16:58' prior: 0!
A number of tests for islands.!

!IslandTests methodsFor: 'tests-simple' stamp: 'ar 5/22/2005 13:39'!
actualClassOf: obj
	^obj send: #class to: obj! !

!IslandTests methodsFor: 'assertions' stamp: 'mtf 6/2/2010 11:02'!
assertIslandIs: anIsland
	self assert: Processor activeIsland == anIsland.
	^ 24! !

!IslandTests methodsFor: 'assertions' stamp: 'mtf 6/2/2010 13:38'!
assertIslandIs: anIsland andSignal: anException
	^ [[
		self assertIslandIs: anIsland.
		anException signal.
		self assertIslandIs: anIsland.
		24
	] on: anException do: [:ex |
		self assertIslandIs: anIsland.
		ex pass]
	] on: UnhandledError do: [:ex |
		self assertIslandIs: anIsland.
		ex pass]! !

!IslandTests methodsFor: 'assertions' stamp: 'mtf 6/2/2010 13:34'!
assertIslandIs: anIsland andSignalChecking: anException
	^ [[
		self assertIslandIs: anIsland.
		self assert: anException signal = 42.
		self assertIslandIs: anIsland.
		24
	] on: anException do: [:ex |
		self assertIslandIs: anIsland.
		ex pass]
	] on: UnhandledError do: [:ex |
		self assertIslandIs: anIsland.
		ex pass]! !

!IslandTests methodsFor: 'tests-simple' stamp: 'ar 5/22/2005 13:57'!
testBitBlt
	| island thisForm thatForm |
	thisForm := Form extent: 100@100 depth: 8.
	island := Island new.
	thatForm := island new: Form.
	thatForm setExtent: 100@100 depth: 8.
	self shouldnt:[thisForm displayOn: thatForm] raise: Error.
	self shouldnt:[thatForm displayOn: thisForm] raise: Error.
	self shouldnt:[thisForm getCanvas drawImage: thatForm at: 0@0] raise: Error.
	self shouldnt:[thatForm getCanvas drawImage: thisForm at: 0@0] raise: Error.! !

!IslandTests methodsFor: 'tests-simple' stamp: 'hrs 10/8/2009 15:58'!
testImplicitIsland
	| island obj |
	island := Island named: 'test'.
	Smalltalk at: #CWindow ifPresent:[:cls|
		obj := island new: cls.
		self assert: obj island == island.
		[
		self assert: obj costume island == island.
		self assert: obj costume frontPage island == island.
		"...etc..."
		] on: CUnsynchronizedModification do:[:ex| ex resume].
	].
	Smalltalk at: #SystemWindow ifPresent:[:cls|
		obj := island new: cls.
		self assert: obj island == island.
		self assert: obj firstSubmorph island == island.
		"... etc ..."
	].! !

!IslandTests methodsFor: 'tests-simple' stamp: 'ar 1/17/2005 23:01'!
testIslandGlobal
	| island obj |
	island := Island named: 'test'.
	obj := island new: TestVariable.
	island at: #test put: obj.
	self assert: (island at: #test) == obj.
! !

!IslandTests methodsFor: 'tests-simple' stamp: 'ar 1/17/2005 23:01'!
testIslandNew
	| island obj |
	island := Island named: 'test'.
	obj := island new: TestVariable.
	self assert: (obj isFarRef).
	self assert: (obj island == island).! !

!IslandTests methodsFor: 'tests-simple' stamp: 'ar 1/17/2005 23:01'!
testIslandValue
	| island obj |
	island := Island named: 'test'.
	obj := island new: TestVariable.
	self assert:(obj value == nil).
	obj value: 42.
	self assert: (obj value == 42).
	obj value: island.
	self assert: (obj value == island).
! !

!IslandTests methodsFor: 'tests-simple' stamp: 'ar 1/17/2005 22:59'!
testIslandVeil
	| island |
	island := Island named: 'test'.
	self assert: (island isFarRef).! !

!IslandTests methodsFor: 'tests-simple' stamp: 'ar 1/17/2005 23:01'!
testNearFarRef
	| island var |
	island := Island named: 'test'.
	var := island new: TestVariable.
	self assert: (self island classOf: var) == FarRef.
	self assert: (island classOf: var) == TestVariable.
	var := TestVariable new.
	self assert: (self island classOf: var) == TestVariable.
	self assert: (island classOf: var) == FarRef.

! !

!IslandTests methodsFor: 'tests-simple' stamp: 'ar 1/17/2005 23:01'!
testReferentialIdentity
	| island var obj |
	island := Island named: 'test'.
	var := island new: TestVariable.
	island at: #test put: var.
	obj := island at: #test.
	self assert: (var class == FarRef).
	self assert: (obj class == FarRef).
	self assert: (var == obj).

	var := TestVariable new.
	island at: #test put: var.
	obj := island at: #test.
	self assert: (var class == TestVariable).
	self assert: (obj class == TestVariable).
	self assert: (var == obj).

! !

!IslandTests methodsFor: 'tests-simple' stamp: 'mtf 6/2/2010 11:36'!
testSyncSend
	| remote localIsland remoteIsland |
	localIsland := Processor activeIsland.
	remoteIsland := Island new.
	remote := remoteIsland new: self class.
	self deny: localIsland isFarRef.
	self assert: remoteIsland isFarRef.
	self assert: remote isFarRef.

	self assert: 24 = (self assertIslandIs: localIsland).
	self assert: 24 = (remote assertIslandIs: remoteIsland).
	self assert: 24 = (remote assertIslandIs: remoteIsland andSignal: Notification).

	self assert: 24 = ([remote assertIslandIs: remoteIsland andSignalChecking: Error]
		on: Error do: [:ex |
			self assertIslandIs: localIsland.
			ex resumeUnchecked: 42]).
	
	self assertIslandIs: localIsland.

	self assert: 24 = ([remote assertIslandIs: remoteIsland andSignalChecking: Notification]
		on: Notification do: [:ex |
			self assertIslandIs: localIsland.
			ex resume: 42]).
	
	self assertIslandIs: localIsland.

	self assert: 12 = ([remote assertIslandIs: remoteIsland andSignalChecking: Error]
		on: Error do: [:ex |
			self assertIslandIs: localIsland.
			12]).
	
	self assertIslandIs: localIsland.

	self assert: 12 = ([remote assertIslandIs: remoteIsland andSignalChecking: Error]
		on: Error do: [:ex |
			self assertIslandIs: localIsland.
			ex retryUsing: [12]]).
	
	self assertIslandIs: localIsland.

	self assert: 24 = ([[remote assertIslandIs: remoteIsland andSignalChecking: Error]
		on: Error do: [:ex |
			self assertIslandIs: localIsland.
			UnhandledError signalForException: ex]]
		on: UnhandledError do: [:ex | 
			self assertIslandIs: localIsland.
			ex exception resumeUnchecked: 42]).
	
	self assertIslandIs: localIsland.

! !

ReadWriteStream subclass: #IslandStream
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Islands-Files'!

!IslandStream methodsFor: 'accessing' stamp: 'ar 6/16/2005 19:31'!
nextInt32
	"Read a 32-bit signed integer from the next 4 bytes"
	^(self next: 4) longAt: 1 bigEndian: true! !

!IslandStream methodsFor: 'accessing' stamp: 'ar 6/16/2005 19:31'!
nextInt32Put: int32
	"Write a signed integer to the next 4 bytes"
	^self nextPutAll: ((ByteArray new: 4) longAt: 1 put: int32 bigEndian: true; yourself).! !

!IslandStream methodsFor: 'accessing' stamp: 'ar 6/16/2005 19:31'!
nextUInt32
	"Read a 32-bit signed integer from the next 4 bytes"
	^(self next: 4) unsignedLongAt: 1 bigEndian: true! !

!IslandStream methodsFor: 'accessing' stamp: 'ar 6/16/2005 19:31'!
nextUInt32Put: int32
	"Write a signed integer to the next 4 bytes"
	^self nextPutAll: ((ByteArray new: 4) unsignedLongAt: 1 put: int32 bigEndian: true; yourself).! !

!IslandStream methodsFor: 'accessing' stamp: 'ar 6/16/2005 19:31'!
nextUUID
	| uuid |
	uuid := UUID nilUUID.
	self next: 16 into: uuid.
	^uuid! !

!IslandStream methodsFor: 'accessing' stamp: 'ar 6/16/2005 19:31'!
nextWordsInto: wordArray
	| buffer max blt start width |
	(wordArray class isWords or:[wordArray byteSize = 4]) 
		ifFalse:[^super nextWordsInto: wordArray].

	max := wordArray basicSize*4.
	buffer := ByteArray new: (max min: 4096).
	blt := BitBlt toForm: (Form new hackRow: wordArray).
	blt sourceForm: (Form new hackRow: buffer).
	blt combinationRule: Form over.
	blt sourceX: 0; sourceY: 0; width: 0; height: 1.
	blt destX: 0; destY: 0.
	start := 0.
	[start < max] whileTrue:[
		width := (max - start) min: buffer size.
		self next: width into: buffer startingAt: 1.
		blt destX: start; width: width.
		blt copyBits.
		start := start + width
	].
	wordArray restoreEndianness. "swap if necessary"
	^wordArray! !

!IslandStream methodsFor: 'accessing' stamp: 'ar 6/16/2005 19:31'!
nextWordsPutAll: wordArray
	| buffer max blt start width |
	(wordArray class isWords or:[wordArray byteSize = 4]) 
		ifFalse:[^super nextWordsPutAll: wordArray].

	wordArray restoreEndianness. "swap if necessary"
	max := wordArray basicSize*4.
	buffer := ByteArray new: (max min: 4096).
	blt := BitBlt toForm: (Form new hackRow: buffer).
	blt sourceForm: (Form new hackRow: wordArray).
	blt combinationRule: Form over.
	blt sourceX: 0; sourceY: 0; width: 0; height: 1.
	blt destX: 0; destY: 0.
	start := 0.
	[start < max] whileTrue:[
		width := (max - start) min: buffer size.
		blt sourceX: start; width: width.
		blt copyBits.
		self next: width putAll: buffer.
		start := start + width
	].
	wordArray restoreEndianness. "swap if necessary"
! !

Object subclass: #FarRefMap
	instanceVariableNames: 'array tally flag island mutex'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Islands-Kernel'!

!FarRefMap commentStamp: 'ar 1/17/2005 16:57' prior: 0!
FarRefMap maps objects to their associated far refs. It is essentially a mixture of dictionary and set. 

[**** TBD: Fix this so that we're not having the array be roots for almost all the refs being stored in it (it defeats the point of using weak refs after all) ****]
[**** TBD: Fix this so we're not endlessly growing the array ****]

!

!FarRefMap class methodsFor: 'instance creation' stamp: 'ar 1/14/2005 23:09'!
new
	^self basicNew initialize! !

!FarRefMap methodsFor: 'accessing' stamp: 'eem 6/11/2008 17:35'!
add: farRef
	"Include farRef as one of the receiver's elements, but only if
	not already present. Answer farRef."

	farRef ifNil: [self error: 'Sets cannot meaningfully contain nil as an element'].
	mutex critical: [| index |
		index := self findElementOrFlag: (farRef valueOn: island).
		(array at: index) == flag ifTrue:[self atNewIndex: index put: farRef].
		].
	^farRef! !

!FarRefMap methodsFor: 'accessing' stamp: 'ar 1/16/2005 20:19'!
asArray
	^Array streamContents:[:s| self do:[:ref| s nextPut: ref]].! !

!FarRefMap methodsFor: 'accessing' stamp: 'ar 1/14/2005 23:13'!
at: key 
	"Answer the value associated with the key."

	^ self at: key ifAbsent:[nil]! !

!FarRefMap methodsFor: 'accessing' stamp: 'hrs 1/5/2007 21:11'!
at: key ifAbsent: aBlock 
	"Answer the value associated with the key or, if key isn't found,
	answer the result of evaluating aBlock."
	| element |
	mutex critical: [
		element := array at: (self findElementOrFlag: key).
		element == flag ifTrue:[^ aBlock value].
		].
	^element! !

!FarRefMap methodsFor: 'accessing' stamp: 'ar 1/14/2005 23:32'!
at: key ifAbsentPut: aBlock
	"Return the value at the given key. 
	If key is not included in the receiver store the result 
	of evaluating aBlock as new value."
	^self at: key ifAbsent:[self at: key put: aBlock value]! !

!FarRefMap methodsFor: 'accessing' stamp: 'ar 1/17/2005 16:39'!
at: aKey put: farRef
	farRef privateValue: aKey island: island.
	^self add: farRef.! !

!FarRefMap methodsFor: 'private' stamp: 'ar 5/22/2005 14:34'!
atNewIndex: index put: anObject
	array at: index put: anObject.
	tally := tally + 1.
	self fullCheck.! !

!FarRefMap methodsFor: 'accessing' stamp: 'eem 6/11/2008 17:35'!
do: aBlock

	tally = 0 ifTrue:[^self].
	mutex critical: [| obj |
		1 to: array size do:[:i|
			((obj := array at: i) == nil or:[obj == flag]) ifFalse:[aBlock value: obj].
		].
	].! !

!FarRefMap methodsFor: 'private' stamp: 'eem 6/11/2008 17:35'!
finalizeValues

	mutex critical: [
		| length oldIndex newIndex element |
		length := array size.
		1 to: length do:[:index|
			(array at: index) ifNil:[
				tally := tally - 1.
				oldIndex := index.
				[oldIndex = length
					ifTrue: [oldIndex := 1]
					ifFalse: [oldIndex := oldIndex + 1].
				(element := array at: oldIndex) == flag] whileFalse:[
					newIndex := self findElementOrFlag: element.
					oldIndex = newIndex ifFalse: [array swap: oldIndex with: newIndex].
				].
			].
		].
	].! !

!FarRefMap methodsFor: 'private' stamp: 'ar 1/10/2007 11:21'!
findElementOrFlag: anObject
	"Answer the index of a first slot containing either a nil (indicating an empty slot) or an element that matches the given object. Answer the index of that slot or zero. Fail if neither a match nor an empty slot is found."

	| index |
	[(index := self scanFor: anObject) = 0] whileTrue:[
		self grow.
	].
	^index! !

!FarRefMap methodsFor: 'private' stamp: 'ar 5/12/2005 18:01'!
fullCheck
	"Keep array at least 1/4 free for decent hash behavior"
	| minFree nowFree |
	minFree := array size // 4 max: 1.
	nowFree := array size - tally.
	nowFree > minFree ifTrue:[^self].
	"Make some room by dumping niled out elements"
	self growTo: array size.
	nowFree := array size - tally.
	nowFree > minFree ifTrue:[
		"Try to shrink if we have very items"
		tally < minFree ifTrue:[self shrink].
		^self].
	"Really grow"
	self grow.! !

!FarRefMap methodsFor: 'private' stamp: 'ar 1/14/2005 23:04'!
grow
	"Grow the elements array and reinsert the old elements"

	self growTo: array size + self growSize! !

!FarRefMap methodsFor: 'private' stamp: 'ar 1/14/2005 23:09'!
growSize
	^ array size max: 2! !

!FarRefMap methodsFor: 'private' stamp: 'eem 6/11/2008 17:36'!
growTo: anInteger
	"Grow the elements array and reinsert the old elements"

	mutex critical: [
		| oldElements |
		oldElements := array.
		array := WeakArray new: anInteger.
		array atAllPut: flag.
		tally := 0.
		oldElements do:[:obj | (obj == flag or: [obj == nil]) ifFalse: [
			array at: (self findElementOrFlag: (obj valueOn: island)) put: obj.
			tally := tally + 1.
			]].
		].! !

!FarRefMap methodsFor: 'initialize' stamp: 'bgf 3/21/2009 21:16'!
initialize
	flag := Object new.
	tally := 0.
	mutex := Mutex new.
	array := WeakArray new: (Set sizeFor: 40).
	array atAllPut: flag.
! !

!FarRefMap methodsFor: 'accessing' stamp: 'ar 1/14/2005 23:16'!
island
	^island! !

!FarRefMap methodsFor: 'accessing' stamp: 'ar 1/14/2005 23:16'!
island: anIsland
	island := anIsland! !

!FarRefMap methodsFor: 'private' stamp: 'ar 5/14/2005 21:38'!
rehash
	self growTo: array size.! !

!FarRefMap methodsFor: 'private' stamp: 'ar 1/16/2005 20:31'!
scanFor: anObject
	"Scan the key array for the first slot containing either a nil (indicating an empty slot) or an element that matches anObject. Answer the index of that slot or zero if no slot is found. This method will be overridden in various subclasses that have different interpretations for matching elements."
	| finish hash start element |
	finish := array size.
	finish > 4096
		ifTrue: [hash := anObject identityHash * (finish // 4096)]
		ifFalse: [hash := anObject identityHash].
	start := (hash \\ array size) + 1.

	"Search from (hash mod size) to the end."
	start to: finish do:[:index | 
		(element := array at: index) ifNotNil:[
			(element == flag or: [(element valueOn: island) == anObject])
				ifTrue: [^ index ]]].

	"Search from 1 to where we started."
	1 to: start-1 do:[:index | 
		(element := array at: index) ifNotNil:[
			(element == flag or: [(element valueOn: island) == anObject])
				ifTrue: [^ index ]]].

	^ 0  "No match AND no empty slot"! !

!FarRefMap methodsFor: 'private' stamp: 'ar 1/16/2005 22:25'!
shrink
	"Shrink the elements array and reinsert the old elements"
	self growTo: (Set sizeFor: tally)! !

!FarRefMap methodsFor: 'accessing' stamp: 'ar 1/14/2005 23:13'!
size
	"Be warned - size is an estimate"
	^tally! !

Object subclass: #Island
	instanceVariableNames: 'myEventMap myScripts name uid globals exports randomStream random'
	classVariableNames: 'Default'
	poolDictionaries: ''
	category: 'Islands-Kernel'!

!Island commentStamp: 'ar 1/17/2005 21:21' prior: 0!
Instance variables:
	name 	<String>	The name of this island.
	uid 		<UUID>	The uid of this island.
	globals	<Dictionary>	The named entry points of this island.
	exports	<FarRefMap>	All the exported references from this island.

Documentation:

Islands
========

An island is a unit of isolation for sets of interacting objects. Any object lives on precisely one island and never "leaves" its island. Put differently, no two objects on different islands can refer to each other directly. Instead, objects on different islands refer to each other via so-called "far references" (FarRef) which are well-known objects controlling and relaying the incoming messages to its designated receiver on the correct island.

Installing Islands
--------------------

At this point, islands need to be manually installed since the work isn't finished yet and some optimization are incomplete (leading to significant slowdowns and other odd effects which may be too much for to deal with when you're just trying to get work done).

To install islands execute:

		Island install.

To uninstall islands execute:

		Island uninstall.

Uninstalling islands will remove all previously created islands but should otherwise have no bad effects on your environment (unless, of course, you are playing around inside the islands code itself; then all bets are off ;-) It is therefore safe to install islands, and if something bad happens (and you still have access to a workspace) uninstall them afterwards to restore the system into a functioning state.

Using Islands
---------------

When the islands have been installed, you are already using them since per definition all objects live in an island.

[	"Example: Printing the current island"
	Transcript cr; show: 'We are currently on ', self island printString.
]

By default all objects are created in the current island. If an object needs to be created on a specific island, one can use the #new: message.


[	"Example: Creating an island and an object on the island"
	| island var |
	island := Island named: 'example'.
	var := island new: TestVariable.
	Transcript cr; show: 'This is a ', var printString.
	Transcript cr; show: 'It lives on ', var island printString.
]

The above will print as "Far:[a TestVariable(nil)]" where "Far[...]" indicates that the object is not living on the current island. And although the object is living on a different island, we can send messages to it, for example:

[	"Example: Sending messages to objects on other islands"
	| island var |
	island := Island named: 'example'.
	var := island new: TestVariable.
	var value: 42.
	Transcript cr; show: 'The value of ', var printString,' is now: ', var value.
].

When references to objects are passed forth and back they are converted from and to far references as needed so that an object which is local to the current island remains local even if this reference is passed to another island and (directly or indirectly) returned to "its" island.

[	"Example: Converting between near and far ref"
	| island var |
	island := Island named: 'example'.
	var := island new: TestVariable.

	"Print my point of view"
	Transcript cr; show: 'From here it looks like a '.
	Transcript show: (self island classOf: var).

	"Print the other point of view"
	Transcript cr; show: 'From there it looks like a '.
	Transcript show: (island classOf: var).

	"And the other way around"
	var := TestVariable new. "created in the current island"

	"Print my point of view"
	Transcript cr; show: 'From here it looks like a '.
	Transcript show: (self island classOf: var).

	"Print the other point of view"
	Transcript cr; show: 'From there it looks like a '.
	Transcript show: (island classOf: var).

]

Far refs also preserve identity, e.g., from the point of view of each island there is a unique identity for all objects (though if it is a far reference or not depends on whether the object has been allocated on the island).

[	"Example: Referential identity"
	| island var obj |
	island := Island named: 'example'.
	var := island new: TestVariable.

	"First use a far ref for testing"
	island at: #test put: var.	"set on its island"
	obj := island at: #test. 		"retrieve from its island"
	Transcript cr; show: 'The classes are: ', var class name, ' and ', obj class name.
	Transcript cr; show: 'Are they identical? ', (var == obj) printString.

	"Now use an object which is far on the other island"
	var := TestVariable new. 	"on our island"
	island at: #test put: var.	"store on other island"
	obj := island at: #test.		"retrieve from other island"
	Transcript cr; show: 'The classes are: ', var class name, ' and ', obj class name.
	Transcript cr; show: 'Are they identical? ', (var == obj) printString.

]

We can always ask any object which island it lives on and this of course extends to objects which were implicitly created while executing code on some island.

[	"Example: Printing an object's island"
	| island window |
	island := Island named: 'example'.
	"Create a complex object"
	window := island new: CWindow.
	Transcript cr; show: 'My island is: ', self island.
	Transcript cr; show: 'The window lives on: ', window island.
	Transcript cr; show: 'And its front page: ', window costume frontPage island.
]

[Side bar: Note that because of the implicit creation rule there can be a huge difference between creating an object yourself or having the other object create it. For example, consider the following:

	Object>>inspect
		"Open an inspector on the object"
		^self inspectorClass openOn: self.

	Object inspectorClass
		"Answer the class that should be used to inspect instances of the receiver"
		^Inspector

When we have an object from a different island, executing "object inspect" will create the inspector on the island the object lives on, whereas "object inspectorClass openOn: object" will create the inspector in the island of the sender. As you may expect, the entire system is currently very lazy in this regard and is therefore also quite inefficient]

Passing objects between islands
------------------------------------

So far, we have assumed that all objects live on precisely one island and are referenced exclusively via far refs when they are used from other islands. Sometimes this is awkward since there is some overhead involved in sending messages between the islands and it can be advantageous to simply copy the object to the other island (not preserving any local reference to it).

When objects are being passed between islands, their classes are queried about how they would like their instances to be passed along to another island. This message is called #howToPassAsArgument (and yes, it is a stupid name but I couldn't come up with any better) and the possible responses are:

- #passByProxy:
	Pass the object by converting it from and to far reference (this is the default)

- #passByCopy:
	Pass the object by copying it. All objects referred to by the copied object will be passed depending on their own strategy. For example, an array containing a passByProxy: object would convert this object into a far reference.

- #passByClone:
	Pass the object by cloning it. This is like passByCopy: except that the object may be copied multiple times. passByClone: is more efficient than passByCopy: but has the requirement that the object may not refer back to itself, neither directly nor indirectly via other passByCopy: or passByClone: objects (if it does, infinite recursion will occur).

- #passByIdentity:
	Pass the object along without doing anything to it. This strategy must only be used under very specific circumstances (see below).

[**** TBD: I think we need at least the ability to explicitly pass an object by reference or by copy if a client wishes so. E.g., what I'd like to see here is something like:
	sharedData := Array new: 100.
	farRef fill: sharedData asPassByProxy. 
and possibly the same for asPassByCopy so that we can copy an object just once ****]

More about passByIdentity:
-------------------------------

The passByIdentity: strategy must be used with great care since it allows islands to access these objects concurrently and has the potential to leak references between islands. A "normal user" should never use passByIdentity: but if you have to you must be aware of the following: The rule of thumb is that it is only safe to use passByIdentity: if an object is both immutable and not containing pointers. Because of its immutability there is no problem with concurrently executing code from different islands; because it doesn't contain pointers it cannot leak them. Conceptually, these objects live on all islands at once - you cannot ask (for example) the number 1 which island it is living on (it will just answer the current island).

It can still be safe to have pointers if the object is immutable. In this situation one needs to store a far reference if the object stored is anything but passByIdentity: itself and resolve the value when it is being read - this strategy is used for example by IslandVariables which are passByIdentity: (they have to be since they are referred to from compiled methods and may therefore be implicitly used from different islands).

If the object is both, mutable as well as (potentially) containing pointers one must be aware that these objects can be read and written to concurrently and have to be specifically designed for this task. The only objects for which this is currently true are IslandVariables and they should generally be used as a template for how to make proper passByIdentity: objects if that is needed.

At this point there are also numerous other objects which are passByIdentity: (browse the references to passByIdentity: to find them) and which do not adhere to the above rules. Most importantly, all Classes are passByIdentity: (since otherwise the VM would break) but also certain process-related objects (Process, ProcessorScheduler, ScriptScheduler, Semaphore, Delay) and certain rendering-related objects (BitBlt, StrikeFont, Canvas).

All of these will be cleaned up over time; right now one has to be careful not to abuse any of them.

Checkpointing
----------------

Islands can be checkpointed and restored using the (-->)image segment technology. These checkpoints can be used to create a copy (or mutiple copies) of the checkpointed island.

[	"Example: Checkpointing and restoring an island"
	| island other varA varB data |
	island := Island named: 'example'.
	varA := island new: TestVariable.
	varA value: 'Hello World'.
	island at: #var put: varA. "or else we cannot retrieve it later"

	"Now checkpoint (VERY slow right now)"
	data := Island checkpoint: island.

	"And restore (VERY fast)"
	other := Island resurrect: data.
	varB := other at: #var.

	Transcript cr; show: 'The value of the restored variable is: ', varB value printString.
]

When using the image segment technology it is important to understand about what is known as "out pointers". With islands, "out pointers" are all objects referred to both from the current island and any other island ("shared references"). 

These shared references will include two types of objects: a) passByIdentity: objects which are referred to from the current and some other island and b) far references which refer to the objects being used from other islands. In addition, clients must expect to find all compiler literals in the shared references even if they aren't passByIdentity: (including arrays and strings).

It is the client's responsibility to do something sensible with these shared references, for example deal with how to store passByIdentity: objects on files or how to store/resolve far references to other islands.

[**** TBD: Make it possible for clients to look at far references BEFORE checkpointing so that an informed decision can be made about what to do with them before even attempting to checkpoint. ****]


Static variables (globals, class vars, pools):
---------------------------------------------------

Islands have support for static variables by using IslandVariables instead of associations for class, pool, and global variables. There are however a number of significant drawbacks. In short, the use of static variables should be avoided if at all possible. If they cannot be avoided they should be used for passByIdentity: objects only (numeric constants for example) since those have the least overhead; or alternatively for passByProxy: objects (more overhead than passByIdentity: but still okay). Objects using passByCopy: (or passByClone:) should be avoided since they will indeed force the contents to be copied.

[	"Example: Comparing read/write speed for globals"
	| varA varB timeA timeB count ref copy print |
	timeA := timeB := 0.
	ref := self island. "any passByProxy: object"
	copy := 'Hello World'. "any passByCopy: object -- size matters!!!!!!"
	varA := Association new. "standard assoc"
	varB := IslandVariable new. "island static var"

	print :=[:label| 
		Transcript cr; show: label.
		Transcript show: ' ', timeA, ' ms versus ', timeB, ' ms'.
		Transcript show: ' (', (timeB asFloat / timeA truncateTo: 0.01), ' x slower)'.
	].

	"Writing passByIdentity: objects (2x slower)"
	count := 1000000. Smalltalk garbageCollect.
	timeA := [1 to: count do:[:i| varA value: i]] timeToRun.
	timeB := [1 to: count do:[:i| varB value: i]] timeToRun.
	print value: 'Writing passByIdentity:'. 

	"Reading passByIdentity: objects (3x slower)"
	count := 1000000. Smalltalk garbageCollect.
	timeA := [1 to: count do:[:i| varA value]] timeToRun.
	timeB := [1 to: count do:[:i| varB value]] timeToRun.
	print value: 'Reading passByIdentity:'. 

	"Writing passByProxy: objects (16x slower)"
	count := 100000. Smalltalk garbageCollect.
	timeA := [1 to: count do:[:i| varA value: ref]] timeToRun.
	timeB := [1 to: count do:[:i| varB value: ref]] timeToRun.
	print value: 'Writing passByProxy:'. 

	"Reading passByProxy: objects (10x slower)"
	count := 100000. Smalltalk garbageCollect.
	timeA := [1 to: count do:[:i| varA value]] timeToRun.
	timeB := [1 to: count do:[:i| varB value]] timeToRun.
	print value: 'Reading passByProxy:'. 

	"Writing passByCopy: objects (16x slower)"
	count := 100000. Smalltalk garbageCollect.
	timeA := [1 to: count do:[:i| varA value: copy]] timeToRun.
	timeB := [1 to: count do:[:i| varB value: copy]] timeToRun.
	print value: 'Writing passByCopy:'. 

	"Reading passByCopy: objects ( >100 x slower)"
		"Note: This needs a little hack to simulate reading a global
		from another island. As long as we remain on the same island
		reading is at passByProxy: speed."
		varB value: (Island new asFarRef: copy).
	count := 100000. Smalltalk garbageCollect.
	timeA := [1 to: count do:[:i| varA value]] timeToRun.
	timeB := [1 to: count do:[:i| varB value]] timeToRun.
	print value: 'Reading passByCopy:'. 
]

Incoming message control
------------------------------

Islands can control how they wish to process incoming messages from other islands, e.g., whether they want to respond synchronously or asynchronously, whether to perform message replication of incoming messages etc. To control how incoming messages are processed one needs to create a subclass of both Island and FarRef (say MyIsland and MyFarRef); implement MyIsland>>asFarRef: to answer a MyFarRef which (in its doesNotUnderstand: handler) provides the appropriate mechanism for incoming messages.

[**** TBD: Think about if (and if so how) we would control passing far refs forth and back between third parties. It is possible that an island might hand out different far refs to different islands (this would allow numerous optimizations between well-known kinds of islands) but then we may need to go back to the original island and ask it about a far ref for a specific island. 
	FarRef>>valueOn: and FarRef>>varValueOn: would be the likely places to handle this - we could store the dstIsland here and if it is 'wrong' we'd just ask our island for the correct reference. But this may also need support from FarRefMap etc. ****]

Merging Islands
------------------
TBD. Merging islands is great to create objects from checkpoints which is MUCH faster than creating them from code.

[	"Example: Creating an object from code vs. from checkpoint"
	| island obj data timeA timeB|
	island := Island named: 'example'.
	Smalltalk garbageCollect.
	timeA := [obj := island new: Mines] timeToRun.			"1800 ms"
	island at: #root put: obj.
	data := Island checkpoint: island.
	Smalltalk garbageCollect.
	timeB := [island := Island resurrect: data] timeToRun.	"7 ms (!!!!!!)"
	"(island at: #root) openInWorld." "and yes it works"
	Transcript cr; show: 'Time to create Mines from code: ', timeA.
	Transcript cr; show: 'Time to create Mines from checkpoint: ', timeB.
]

However, using this mechanism for creating objects requires us to be able to merge an island with the island where we'd like the object to be present so this would need to be addressed (and is tricky to the point that it might be faster to just run the code ;-)

Optimizing islands (tips and tricks)
-----------------------------------------

TBD. State the rules for optimizing islands. So far we know:

* Avoid globals like mad except for passByIdenity: - even passByProxy: can be fairly problematic if many objects are being written since they force far refs to be created which has a significant impact on GC (such as Morphic's use of ActiveEvent and friends). Worst case examples are using static arrays of arrays such as used by chess game.

* Avoid sending lots of short-lived passByProxy: objects to other islands (MorphicEvents for example) but rather make make them passByCopy: - managing far refs can be more expensive than copying the object (for small, short-lived objects)

* Avoid extreme multi-dispatching between objects on different islands, e.g., having Morph>>fullDrawOn: call Canvas>>fullDraw: call Morph>>fullDrawOn: etc. can be a killer if the objects live on different islands. For complex operations (such as drawing) consider copying on the demand or caching to avoid round trips.

* Carefully think about who is responsible for creating the object - using the debug halo handle on a Morph created on another island currently creates an inspector on THAT island which means both extra round trips for display/interaction as well as increasing that island (project) size.
!

!Island class methodsFor: 'accessing' stamp: 'ar 5/22/2005 15:42'!
default
	"Return the system-wide default island"
	| island |
	Default ifNil:[
		Default := self basicNew initialize.
		Default name: 'Squeak' copy.
	].
	island := Processor activeProcess island.
	island ifNil:[^Default] ifNotNil:[^(Default asFarRef: Default) valueOn: island].! !

!Island class methodsFor: 'system startup' stamp: 'mtf 1/28/2011 14:27'!
initialize
	"Island initialize"
	Smalltalk addToStartUpList: self.! !

!Island class methodsFor: 'instance creation' stamp: 'ar 1/13/2005 22:16'!
named: aString
	^self new name: aString! !

!Island class methodsFor: 'instance creation' stamp: 'ar 3/23/2006 16:03'!
new
	"Create a new island."
	| island prior |
	prior := Processor activeIsland.
	island := self basicNew.
	Processor activeIsland: island.
	island initialize.
	Processor activeIsland: prior.
	^island asFarRef: island! !

!Island class methodsFor: 'system startup' stamp: 'mtf 1/28/2011 14:45'!
startUp: resuming
"on resume from disk, reinitialize the random streams, so that two copies of the image don't have identical random streams. I believe this is correct for replicated TIslands as well, as a TIsland that survives a save and resume is definitely not connected to a router, synced, or even able to run in the current implementation, I believe"

	resuming ifTrue: [self allSubInstancesDo: [:ea | ea initializeRandomStream]]! !

!Island methodsFor: 'accessing' stamp: 'ar 5/24/2005 16:40'!
asFarRef: anObject
	"Answer the far ref for the given (near) object"
	anObject isFarRef ifTrue:[^anObject].
	^exports at: anObject ifAbsentPut:[
		| ref |
		ref := FarRef new.
		anObject myFarRef: ref.
		ref].! !

!Island methodsFor: 'accessing' stamp: 'ar 3/26/2007 22:46'!
asyncScriptMessageSend 
	"Overwritten in TIsland"
	^AsyncScriptMessageSend! !

!Island methodsFor: 'accessing' stamp: 'ar 3/13/2006 17:07'!
at: globalName
	^globals at: globalName ifAbsent:[nil]! !

!Island methodsFor: 'accessing' stamp: 'ar 1/13/2005 22:16'!
at: globalName ifAbsent: aBlock
	^globals at: globalName ifAbsent: aBlock! !

!Island methodsFor: 'accessing' stamp: 'jcg 1/23/2008 15:47'!
at: globalName ifAbsentPut: aBlock
	^globals at: globalName ifAbsentPut: aBlock! !

!Island methodsFor: 'accessing' stamp: 'sn 2/23/2018 23:36'!
at: globalName put: value

	^globals at: globalName put: value! !

!Island methodsFor: 'utilities' stamp: 'ar 1/17/2005 16:44'!
classOf: anObject
	"Answer the class of anObject"
	^anObject class! !

!Island methodsFor: 'accessing' stamp: 'ar 1/30/2006 23:08'!
controllerID
	^nil! !

!Island methodsFor: 'utilities' stamp: 'tk 6/3/2005 11:55'!
copy: farRef
	"Bring a copy of the value of the ref here, completely resolved to be in the this Island."

	"Must be performed with me as the current island!!"
	^ farRef isFarRef 
		ifTrue: [farRef copyToCurrentIsland]
		ifFalse: [farRef]
! !

!Island methodsFor: 'initialize' stamp: 'ar 6/17/2005 17:06'!
flushGlobals
	globals := IdentityDictionary new.! !

!Island methodsFor: 'accessing' stamp: 'ar 6/17/2005 16:54'!
globalNameAt: value
	^globals keyAtIdentityValue: value! !

!Island methodsFor: 'accessing' stamp: 'sn 2/23/2018 23:20'!
globalNameAt: value ifAbsent: exceptionBlock
	^globals keyAtIdentityValue: value ifAbsent: exceptionBlock.! !

!Island methodsFor: 'accessing' stamp: 'ar 9/8/2005 11:58'!
globals
	^globals! !

!Island methodsFor: 'initialize' stamp: 'mtf 1/28/2011 14:30'!
initialize
	exports := FarRefMap new.
	exports island: self.
	globals := IdentityDictionary new.

"@@@@ FIXME: This needs to go out of here and be moved to TIsland since it has a dependency on the RC4 cypher. After the unification of Island/TIsland this should just go away."
	self initializeRandomStream! !

!Island methodsFor: 'initialize' stamp: 'mtf 1/28/2011 14:29'!
initializeRandomStream
"@@@@ FIXME: This needs to go out of here and be moved to TIsland since it has a dependency on the RC4 cypher. After the unification of Island/TIsland this should just go away."

	| seed |
	random := Random seed: Time totalSeconds.
	Smalltalk at: #TCryptoRandom ifPresent:[:randomClass|
		seed := ByteArray new: 128.
		randomClass gatherEntropy: seed.
		Smalltalk at: #TARC4Cypher ifPresent:[:cypherClass|
			randomStream := cypherClass new.
			randomStream key: seed.
		].
	].! !

!Island methodsFor: 'timing' stamp: 'ar 9/17/2005 21:20'!
millisecondClockValue
	"Answer the value of the millisecond clock counter."
	^RealTime millisecondClockValue! !

!Island methodsFor: 'private' stamp: 'ar 9/15/2005 23:02'!
myEventMap
	^myEventMap! !

!Island methodsFor: 'private' stamp: 'ar 9/15/2005 23:02'!
myEventMap: aMap
	myEventMap := aMap! !

!Island methodsFor: 'private' stamp: 'ar 9/15/2005 23:02'!
myScripts
	^myScripts ifNil:[#()].! !

!Island methodsFor: 'private' stamp: 'ar 9/15/2005 23:02'!
myScripts: scripts
	^myScripts := scripts.! !

!Island methodsFor: 'accessing' stamp: 'ar 1/13/2005 22:16'!
name
	^name! !

!Island methodsFor: 'accessing' stamp: 'ar 1/13/2005 22:16'!
name: aString
	name := aString.! !

!Island methodsFor: 'utilities' stamp: 'ar 1/17/2005 16:46'!
new: aClass
	"Create an instance of aClass"
	^aClass new! !

!Island methodsFor: 'utilities' stamp: 'ar 5/28/2005 20:23'!
new: aClass constructor: aSymbol arguments: args
	"Create an instance of aClass"
	^aClass perform: aSymbol withArguments: args! !

!Island methodsFor: 'initialize' stamp: 'ar 6/15/2009 11:57'!
noteNewProcess: aProcess
	"Inform the receiver that a new process was created carrying this island."
! !

!Island methodsFor: 'accessing' stamp: 'ar 1/17/2005 14:02'!
printOn: aStream
	super printOn: aStream.
	aStream nextPutAll: '('; nextPutAll: (name ifNil:[uid]) asString; nextPutAll: ')'.! !

!Island methodsFor: 'accessing' stamp: 'ar 5/24/2005 16:40'!
project
	^globals at: #project ifAbsent:[nil]! !

!Island methodsFor: 'accessing' stamp: 'ar 5/24/2005 16:40'!
project: aProject
	^globals at: #project put: aProject! !

!Island methodsFor: 'accessing' stamp: 'ar 9/17/2005 12:02'!
random
	^random! !

!Island methodsFor: 'accessing' stamp: 'ar 9/17/2005 12:02'!
random: aRandom
	random := aRandom! !

!Island methodsFor: 'accessing' stamp: 'ar 1/28/2006 21:44'!
randomStream
	^randomStream! !

!Island methodsFor: 'accessing' stamp: 'ar 1/28/2006 21:44'!
randomStream: anObject
	randomStream := anObject! !

!Island methodsFor: 'accessing' stamp: 'ar 1/23/2006 16:49'!
register: anObject name: aName! !

!Island methodsFor: 'accessing' stamp: 'ar 1/22/2006 13:10'!
removeGlobal: name
	^globals removeKey: name ifAbsent:[]! !

!Island methodsFor: 'initialize' stamp: 'ar 5/26/2005 12:35'!
roots
	"Answer the roots for this island"
	| roots |
	self flag: #fixMe.
	"I don't quite understand why I have added the following GC but it does not work without it. I need to review it and either add a lengthy comment why we need it or (even better) add a comment explaining what needs to be fixed so we can get by withou it. There really doesn't seem to be much of a reason for it here."
	Smalltalk garbageCollect.
	roots := Array streamContents:[:s|
		s nextPut: self. "always myself first"
		exports do:[:each| s nextPut: each].
	].
	exports growTo: (Set sizeFor: roots size).
	^roots! !

!Island methodsFor: 'accessing' stamp: 'ar 1/30/2006 22:16'!
scheduleMessage: msgSend future: deltaMSecs! !

!Island methodsFor: 'accessing' stamp: 'ar 5/24/2005 16:40'!
scheduler
	^globals at: #scheduler ifAbsent:[nil]! !

!Island methodsFor: 'accessing' stamp: 'ar 5/24/2005 16:40'!
scheduler: aScheduler
	^globals at: #scheduler put: aScheduler! !

!Island methodsFor: 'utilities' stamp: 'ar 5/21/2005 17:48'!
shallowCopy: obj
	"Create a shallow copy of the argument on this island."
	| class copy value |
	obj isFarRef ifFalse:[^obj shallowCopy].
	class := obj send: #class to: obj.
	copy := self new: class.
	1 to: class instSize do:[:i|
		value := obj send: #instVarAt: withArguments: (Array with: i).
		copy instVarAt: i put: value.
	].
	^copy! !

!Island methodsFor: 'initialize' stamp: 'ar 1/14/2005 20:47'!
shutDown
	"Shut down this island"! !

!Island methodsFor: 'initialize' stamp: 'ar 1/14/2005 20:47'!
startUp
	"Start up this island"! !

!Island methodsFor: 'utilities' stamp: 'tk 6/6/2005 13:27'!
test: aClass
	"Create a complex tast case from aClass"
	^aClass test! !

!Island methodsFor: 'timing' stamp: 'das 10/31/2005 14:43'!
time
	"Answer the value of the millisecond clock counter."
	^RealTime millisecondClockValue! !

!Island methodsFor: 'timing' stamp: 'ar 9/17/2005 21:20'!
totalSeconds
	"Answer the number of seconds expired"
	^RealTime totalSeconds! !

!Island methodsFor: 'accessing' stamp: 'ar 1/13/2005 22:16'!
uid
	^uid ifNil:[
		self uid: UUID new.
		uid]! !

!Island methodsFor: 'accessing' stamp: 'ar 1/14/2005 22:34'!
uid: aUUID
	uid ifNotNil:[^self error: 'Cannot change uid'].
	uid := aUUID.! !

!Island methodsFor: 'copying' stamp: 'tk 6/6/2005 16:57'!
veryDeepCopyWith: aDeepCopier
	"Don't duplicate an island this way.  Point at the original."

	! !

Object subclass: #IslandArgumentCopier
	instanceVariableNames: 'srcIsland dstIsland map'
	classVariableNames: 'CopyStats ProxyStats'
	poolDictionaries: ''
	category: 'Islands-Kernel'!

!IslandArgumentCopier commentStamp: 'ar 1/17/2005 21:20' prior: 0!
IslandArgumentCopier is responsible for converting arguments passed between islands. The following 'styles' of argument passing are currently supported:

#passByProxy: 
	Will pass a (far) proxy to the object (default).

#passByCopy: 
	Will pass a copy of the object.

#passByClone: 
	An optimized version of passByCopy: for objects which do not reference themselves.

#passByIdentity:
	Will pass the object directly. Note that passByIdentity: objects have various restrictions on what they are allowed to do and what not most of which are currently not enforced since we're still in an experimental stadium. Use with extreme caution.!

!IslandArgumentCopier class methodsFor: 'accessing' stamp: 'ar 1/14/2005 20:38'!
copyStats
	"IslandArgumentCopier copyStats sortedCounts"
	^CopyStats! !

!IslandArgumentCopier class methodsFor: 'accessing' stamp: 'ar 1/14/2005 20:38'!
copyStats: aBag
	"
		IslandArgumentCopier copyStats: Bag new.
		IslandArgumentCopier copyStats: nil.
	"
	CopyStats := aBag.! !

!IslandArgumentCopier class methodsFor: 'accessing' stamp: 'bgf 7/15/2009 16:14'!
isPassByIdentity: obj
	"Answer whether this object is a legal #passByIdentity: Object"

	^ (obj class howToPassAsArgument == #passByIdentity:) ! !

!IslandArgumentCopier methodsFor: 'copying' stamp: 'sn 2/22/2018 12:11'!
clone: obj
	| newObj |
	CopyStats ifNotNil:[CopyStats add: obj class].
	obj class isPointers ifFalse:[^obj shallowCopy].
	newObj := obj shallowCopy.
	1 to: newObj class instSize do:[:i|
		newObj instVarAt: i put: (self pass: (obj instVarAt: i)).
	].
	1 to: newObj basicSize do:[:i|
		newObj basicAt: i put: (self pass: (obj basicAt: i)).
	].
	^newObj! !

!IslandArgumentCopier methodsFor: 'copying' stamp: 'sn 2/22/2018 12:12'!
copy: obj
	| newObj |
	CopyStats ifNotNil:[CopyStats add: obj class].
	obj class isPointers ifFalse:[^obj shallowCopy].
	newObj := obj shallowCopy.
	map at: obj put: newObj.
	1 to: newObj class instSize do:[:i|
		newObj instVarAt: i put: (self pass: (obj instVarAt: i)).
	].
	1 to: newObj basicSize do:[:i|
		newObj basicAt: i put: (self pass: (obj basicAt: i)).
	].
	^newObj! !

!IslandArgumentCopier methodsFor: 'initialize' stamp: 'ar 1/14/2005 13:06'!
from: partyA to: partyB
	srcIsland := partyA.
	dstIsland := partyB.! !

!IslandArgumentCopier methodsFor: 'copying' stamp: 'mtf 11/2/2011 09:59'!
pass: obj
	srcIsland == dstIsland ifTrue: [^ obj].
	^self perform: obj class howToPassAsArgument with: obj! !

!IslandArgumentCopier methodsFor: 'copying' stamp: 'ar 1/17/2005 14:31'!
passByClone: obj
	"Pass a clone of the given object"
	^self clone: obj! !

!IslandArgumentCopier methodsFor: 'copying' stamp: 'ar 1/14/2005 13:14'!
passByCopy: obj
	"Pass a copy of the given object"
	map ifNil:[map := IdentityDictionary new].
	^map at: obj ifAbsent:[self copy: obj].
! !

!IslandArgumentCopier methodsFor: 'copying' stamp: 'bgf 7/15/2009 16:12'!
passByIdentity: obj
	"Pass the object itself.
	NOTE: passByIdentity objects have VERY special requirements!!!!!!"
	^ obj! !

!IslandArgumentCopier methodsFor: 'copying' stamp: 'ar 5/28/2005 14:40'!
passByProxy: obj
	"Pass the object using a proxy"
	| ref |
	ProxyStats ifNotNil:[ProxyStats add: obj class].
	obj isFarRef 
		ifTrue:[ref := obj]
		ifFalse:[ref := srcIsland asFarRef: obj].
	^ref valueOn: dstIsland! !

IslandArgumentCopier subclass: #IslandDeepCopier
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Islands-Kernel'!

!IslandDeepCopier commentStamp: '<historical>' prior: 0!
For very deep copy of a FarRef.  A tree of objects is created in the current island with all references to other islands resolved.  Just like IslandArgumentCopier except that no FarRefs are ever generated.  Instead, those objects are copied locally.!

!IslandDeepCopier methodsFor: 'copying' stamp: 'tk 6/6/2005 17:27'!
pass: obj
	"Bring the object here, and do not allow FarRefs"

	| how |
	obj isFarRef ifTrue:[^ super pass: obj].
	how := obj class howToPassAsArgument.
	how = #passByProxy: ifTrue: [how := #passByCopy:].	"resolve all refs"
	^ self perform: how with: obj! !

Object subclass: #IslandReader
	instanceVariableNames: 'archive dataStream splObjects imports segment unblessed progressBar defaultInfo'
	classVariableNames: 'ImportCodes'
	poolDictionaries: ''
	category: 'Islands-Files'!

!IslandReader class methodsFor: 'class initialization' stamp: 'ar 6/17/2005 18:59'!
archiveVersionID
	^self subclassResponsibility! !

!IslandReader class methodsFor: 'class initialization' stamp: 'ar 6/16/2005 23:31'!
initialize
	"IslandReader initialize"
	(Smalltalk includesKey: #ByteString) ifFalse:[
		Smalltalk declare: #ByteString from: Undeclared.
		Smalltalk declare: #WideString from: Undeclared.
		Smalltalk at: #ByteString put: 'hello' class.
	].
	IslandWriter initialize.
	ImportCodes := Array new: 256 withAll: #errorInvalidCode.
	IslandWriter classPool keysAndValuesDo:[:key :code|
		(key beginsWith: 'Code') ifTrue:[
			ImportCodes at: code put: ('import', (key allButFirst: 4)) asSymbol.
		].
	].! !

!IslandReader class methodsFor: 'blessings' stamp: 'ar 6/16/2005 19:29'!
isBlessedVersionID: uuid
	^false! !

!IslandReader class methodsFor: 'instance creation' stamp: 'ar 6/16/2005 19:29'!
new
	^self basicNew initialize! !

!IslandReader class methodsFor: 'class initialization' stamp: 'ar 6/17/2005 19:00'!
objectVersionID
	^self subclassResponsibility! !

!IslandReader methodsFor: 'accessing' stamp: 'ar 6/17/2005 15:50'!
defaultInfo
	^defaultInfo! !

!IslandReader methodsFor: 'accessing' stamp: 'ar 6/17/2005 15:50'!
defaultInfo: aString
	defaultInfo := aString.! !

!IslandReader methodsFor: 'private' stamp: 'sn 2/22/2018 21:17'!
endianness
	"Return which endian kind the incoming segment came from"

	segment class isBits ifFalse:
		["Hope that primitive 98 did the right thing - anyway, we lost information about endianness, so pretend we share the image's endianness."
		^Smalltalk endianness].
	^(segment first bitShift: -24) asCharacter == $d ifTrue: [#big] ifFalse: [#little]! !

!IslandReader methodsFor: 'importing' stamp: 'ar 6/16/2005 19:29'!
errorInvalidCode
	self error:'Invalid code'.
	^nil! !

!IslandReader methodsFor: 'importing' stamp: 'ar 6/16/2005 19:29'!
importArray
	"Export an array"
	| sz array |
	sz := dataStream nextUInt32.
	array := Array new: sz.
	1 to: sz do:[:i| array at: i put: self importNext].
	^array! !

!IslandReader methodsFor: 'importing' stamp: 'ar 6/16/2005 20:02'!
importByteString
	| sz string |
	sz := dataStream nextUInt32.
	string := ByteString new: sz.
	string replaceFrom: 1 to: sz with: (dataStream next: sz) asString startingAt: 1.
	^string! !

!IslandReader methodsFor: 'importing' stamp: 'ar 6/16/2005 19:29'!
importByteSymbol
	^self importByteString asSymbol! !

!IslandReader methodsFor: 'importing' stamp: 'ar 6/16/2005 19:29'!
importCharacter
	^Character value: dataStream nextUInt32! !

!IslandReader methodsFor: 'importing' stamp: 'ar 1/24/2006 18:08'!
importClass
	"Export class/version"
	| className classFormat classShape importedClass classVersionID |
	className := self importNext.
	classFormat := self importNext.
	classShape := self importNext.
	classVersionID := self importNext.
	(IslandReader isBlessedVersionID: classVersionID) ifFalse:[
		self flag: #fixMe. "Should we raise an exception here?"
		unblessed add: className.
	].
	importedClass := Smalltalk at: className ifAbsent:[nil].
	importedClass ifNil:[self error: className, ' does not exist'].

	"Verify class structure"
	importedClass format = classFormat 
		ifFalse:[self error: className, ' has changed class format'].
	importedClass allInstVarNames = classShape
		ifFalse:[self error: className, ' has changed class shape'].
	importedClass classVersionID = classVersionID
		ifFalse:[self error: className, ' has changed class version'].

	^importedClass! !

!IslandReader methodsFor: 'importing' stamp: 'ar 6/16/2005 19:29'!
importColor
	^Color colorFromPixelValue: (dataStream nextUInt32) depth: 32! !

!IslandReader methodsFor: 'importing' stamp: 'NiccoKunzmann 7/8/2014 16:09'!
importConstructor
	"Export an array"
	| constructor object value |
	constructor := self importNext.
	object := self importNext.
	value := constructor importFromIslandReader: object.
	^ value! !

!IslandReader methodsFor: 'importing' stamp: 'ar 6/16/2005 19:29'!
importFalse
	^false! !

!IslandReader methodsFor: 'importing' stamp: 'ar 6/16/2005 20:02'!
importFieldDefinition
	| className fieldName fieldDef |
	className := self importNext.
	fieldName := self importNext.
	fieldDef := (Smalltalk at: className ifAbsent:[^nil]) fieldNamed: fieldName.
	fieldDef ifNil:[self error: 'Where is the field???'].
	^fieldDef! !

!IslandReader methodsFor: 'importing' stamp: 'ar 6/16/2005 19:29'!
importFloat
	| float |
	float := Float new: 2.
	float basicAt: 1 put: dataStream nextUInt32.
	float basicAt: 2 put: dataStream nextUInt32.
	^float! !

!IslandReader methodsFor: 'importing' stamp: 'ar 6/16/2005 19:29'!
importFont
	| family height emphasis |
	family := self importNext.
	height := self importNext.
	emphasis := self importNext.
	^StrikeFont familyName: family size: height emphasized: emphasis! !

!IslandReader methodsFor: 'importing' stamp: 'ar 6/17/2005 18:56'!
importGlobalRef
	^self errorInvalidCode! !

!IslandReader methodsFor: 'importing' stamp: 'ar 6/16/2005 19:29'!
importImageSegment
	| data sz |
	sz := dataStream nextUInt32.
	data := WordArrayForSegment new: sz.
	dataStream nextWordsInto: data.
	^data! !

!IslandReader methodsFor: 'importing' stamp: 'ar 6/16/2005 19:29'!
importLargeNegativeInteger
	| sz int |
	sz := dataStream nextUInt32.
	int := LargeNegativeInteger basicNew: sz.
	int replaceFrom: 1 to: sz with: (dataStream next: sz) startingAt: 1.
	^int! !

!IslandReader methodsFor: 'importing' stamp: 'ar 6/16/2005 19:29'!
importLargePositiveInteger
	| sz int |
	sz := dataStream nextUInt32.
	int := LargePositiveInteger basicNew: sz.
	int replaceFrom: 1 to: sz with: (dataStream next: sz) startingAt: 1.
	^int! !

!IslandReader methodsFor: 'importing' stamp: 'ar 6/16/2005 19:29'!
importMetaclass
	"Export class/version"
	^self importNext class! !

!IslandReader methodsFor: 'importing' stamp: 'ar 6/16/2005 20:03'!
importMethodTrigger
	| className selector trigger |
	className := self importNext.
	selector := self importNext.
	trigger := (Smalltalk at: className ifAbsent:[^nil]) methodTriggerAt: selector.
	trigger ifNil:[self error: 'Where is the trigger???'].
	^trigger! !

!IslandReader methodsFor: 'importing' stamp: 'ar 6/16/2005 19:29'!
importNext
	| code |
	code := dataStream next.
	^self perform: (ImportCodes at: code)! !

!IslandReader methodsFor: 'importing' stamp: 'ar 6/16/2005 19:29'!
importNil
	^nil! !

!IslandReader methodsFor: 'importing' stamp: 'ar 6/16/2005 20:03'!
importPoint
	| x y |
	x := self importNext.
	y := self importNext.
	^x @ y! !

!IslandReader methodsFor: 'importing' stamp: 'ar 6/16/2005 19:29'!
importSmallInteger
	^dataStream nextInt32! !

!IslandReader methodsFor: 'importing' stamp: 'ar 6/16/2005 19:29'!
importSpecial
	^splObjects at: self importNext ifAbsent:[nil]! !

!IslandReader methodsFor: 'importing' stamp: 'ar 6/17/2005 18:56'!
importTForm
	^self errorInvalidCode! !

!IslandReader methodsFor: 'importing' stamp: 'ar 6/16/2005 19:29'!
importTrue
	^true! !

!IslandReader methodsFor: 'importing' stamp: 'ar 6/16/2005 19:29'!
importUUID
	| uuid |
	uuid := UUID basicNew: 16.
	dataStream next: 16 into: uuid.
	^uuid! !

!IslandReader methodsFor: 'importing' stamp: 'ar 6/16/2005 20:03'!
importWideString
	| sz string |
	sz := dataStream nextUInt32.
	string := WideString new: sz.
	dataStream nextWordsInto: string.
	^string! !

!IslandReader methodsFor: 'importing' stamp: 'ar 6/16/2005 19:29'!
importWideSymbol
	^self importWideString asSymbol! !

!IslandReader methodsFor: 'initialize' stamp: 'ar 12/11/2005 00:15'!
initialize
	splObjects := IdentityDictionary new.
	splObjects at: #TextStyle::Default put: TextStyle default.
	TextStyle actualTextStyles keysAndValuesDo:[:key :val|
		splObjects at: ('TextStyle::', key) asSymbol put: val.
	].
	Cursor classPool keysAndValuesDo:[:key :val|
		(val isKindOf: Cursor) ifTrue:[
			splObjects at: ('Cursor::', key) asSymbol put: val.
		].
	].
	Smalltalk at: #CCursorLibrary ifPresent:[:aClass|
		aClass default keysAndValuesDo:[:key :val|
			splObjects at: ('CCursorLibrary::', key) asSymbol put: val.
		].
	].
	Smalltalk at: #CIconLibrary ifPresent:[:aClass|
		aClass default icons keysAndValuesDo:[:key :val|
			splObjects at: ('CIconLibrary::Default::', key) asSymbol put: val.
		].
	].
	Smalltalk at: #MenuIcons ifPresent:[:mic|
		(mic classPool at: #Icons) keysAndValuesDo:[:key :val|
			splObjects at: ('MenuIcons::', key) asSymbol put: val.
		].
	].
	Smalltalk at: #ScriptingSystem ifPresent:[:scs|
		scs formDictionary keysAndValuesDo:[:key :val|
			splObjects at: ('ScriptingSystem::Forms::', key) asSymbol put: val.
		].
	].
	Smalltalk at: #CIdentityTransform ifPresent:[:aClass|
		splObjects 
			at: #'CIdentityTransform::Default'
			put: (aClass classPool at: #Default).
	].
	Smalltalk at: #IdentityTransform ifPresent:[:aClass|
		splObjects 
			at: #'IdentityTransform::Default'
			put: (aClass classPool at: #Default).
	].
	unblessed := IdentitySet new.! !

!IslandReader methodsFor: 'private' stamp: 'sn 10/24/2018 21:50'!
loadSegmentFrom: segmentWordArray outPointers: outPointerArray
	"Load segmentWordArray into the memory.  Adapt the primitive to the new API, which is to answer the array of loaded objects, the first of which should be the array of roots.  The primitive will install a binary image segment and return as its value the array
	 of roots of the tree of objects represented.  Upon successful completion, the
	 wordArray will have been becomed into anArray of the loaded objects.  So simply answer the segmentWordArray which will have becommed."

	| segmentFormat |
	segmentFormat := self segmentFormatFrom: segmentWordArray first.
	segmentFormat = Smalltalk imageFormatVersion ifTrue:
		[^(self primitiveLoadSegmentFrom: segmentWordArray outPointers: outPointerArray)
			ifNil: [self error: 'segment load failed']
			ifNotNil: [segmentWordArray]].
	segmentFormat >= 68000
		ifTrue:
			[Smalltalk wordSize = 4 ifTrue:
				[^(Spur64BitImageSegmentLoader new loadSegmentFrom: segmentWordArray outPointers: outPointerArray)]]
		ifFalse:
			[Smalltalk wordSize = 8 ifTrue:
				[^(Spur32BitImageSegmentLoader new loadSegmentFrom: segmentWordArray outPointers: outPointerArray)]].
	self error: 'segment version unrecognized'! !

!IslandReader methodsFor: 'private' stamp: 'sn 10/24/2018 21:49'!
loadSegmentFromData: segmentWordArray outPointers: outPointerArray
	"This primitive will install a binary image segment and return as its value the array of roots of the tree of objects represented.  Upon successful completion, the wordArray will have been transmuted into an object of zero length.  If this primitive should fail, it will have destroyed the contents of the segment wordArray."

		| segmentFormat obj roots lastObject |
		
		segmentFormat := self segmentFormatFrom: segmentWordArray first.
		 segmentFormat = 6502 ifTrue:
			[LegacyImageSegment adoptInstance: self.
			 ^self loadSegmentFrom: segment outPointers: outPointerArray].
		 segmentFormat = Smalltalk imageFormatVersion ifTrue:
			["NativeImageSegment adoptInstance: self."
			"self halt."
				obj := segmentWordArray.
	lastObject := segment nextObject.
	lastObject == 0 ifTrue:[lastObject := Object new].
	roots := self loadSegmentFrom: segmentWordArray outPointers: outPointerArray.
	self startUpFrom: obj to: lastObject.
	^roots at:1
			
			 "^self loadSegmentFrom: segmentWordArray outPointers: outPointerArray"
			].
		 self error: 'no handling for format ', segmentFormat asString. ' in a ', Smalltalk imageFormatVersion asString, ' image.'


	"| obj roots lastObject |
	obj := segmentWordArray.
	lastObject := segment nextObject.
	lastObject == 0 ifTrue:[lastObject := Object new].
	roots := self primitiveLoadSegmentFrom: segmentWordArray outPointers: outPointerArray.
	self startUpFrom: obj to: lastObject.
	^roots"! !

!IslandReader methodsFor: 'private' stamp: 'sn 2/22/2018 22:10'!
primitiveLoadSegmentFrom: segmentWordArray outPointers: outPointerArray
	"This primitive will install a binary image segment and return as its value the array of roots of the tree of objects represented.  Upon successful completion, the wordArray will have been transmuted into an object of zero length.  If this primitive should fail, it will have destroyed the contents of the segment wordArray."

	<primitive: 99 error: ec>	"successful completion returns the array of roots"
	^nil								"failure returns nil"! !

!IslandReader methodsFor: 'accessing' stamp: 'ar 6/17/2005 15:48'!
progressBar
	^progressBar! !

!IslandReader methodsFor: 'accessing' stamp: 'ar 6/17/2005 15:48'!
progressBar: aBlock
	progressBar := aBlock! !

!IslandReader methodsFor: 'reading' stamp: 'sn 10/24/2018 21:49'!
readObjectFrom: aStream
	| versionID data roots |
	Cursor read showWhile:[
		versionID := UUID nilUUID.
		aStream next: 16 into: versionID.
		(versionID = self class archiveVersionID) 
			ifFalse:[^self error: 'Not a project file'].
		archive := ZipArchive new.
		archive readFrom: aStream.
		data := (archive memberNamed: 'object.isg') byteStream contents.
		dataStream := IslandStream with: data.
		dataStream reset.
		(dataStream nextUUID = self class objectVersionID) 
			ifFalse:[^self error: 'Not a croquet file'].
		imports := self importNext.
		segment := self importNext.
	].
	progressBar ifNotNil:[progressBar value: defaultInfo].
	Cursor wait showWhile:[
		roots := self loadSegmentFromData: segment outPointers: imports.
	].
	^roots at: 1! !

!IslandReader methodsFor: 'as yet unclassified' stamp: 'sn 10/24/2018 21:07'!
segmentFormatFrom: a32BitWord
	"The first two words of a segment array contain the image format version of the system upon which the segment was generated, along with a top byte that is either $d or $s (from the 'does' in #doesNotUnderstand:).  But this may be encoded either in big-endian or little-endian format.  Since endianness may or may not have been changed, determining what the segment format is takes care."
	| msc lsc |
	msc := Character value: ((a32BitWord bitShift: -24) bitAnd: 255).
	lsc := Character value: (a32BitWord bitAnd: 255).
	(('ds' includes: msc)
	 and: ['ds' includes: lsc]) ifTrue:
		[self error: 'ambiguous segment format'].
	('ds' includes: msc) ifTrue:
		[^a32BitWord bitAnd: 16rFFFFFF].
	^((a32BitWord bitShift: -24) bitAnd: 16rFF)
	+ ((a32BitWord bitShift: -8) bitAnd: 16rFF00)
	+ ((a32BitWord bitShift: 8) bitAnd: 16rFF0000)! !

!IslandReader methodsFor: 'private' stamp: 'sn 10/24/2018 21:50'!
startUpFrom: firstObj to: lastObj
	"Send startUp messages to all objects"
	| count |
	
	count := 0.
	firstObj do: [:importedObject |
		
		importedObject startUpInIslandReader: self.
		count := count + 1.
		].
		^count
	
	"count := 0.
	obj := firstObj."
	"Note: This order has been carefully chosen so that even if the object #becomes: itself into something different it will work. This is needed for example by MethodDictionary."
	"[
	self halt.next := obj nextObject.
	count := count + 1.
	(obj class = MethodDictionary) ifTrue:[self halt. Transcript show: obj; cr].
	obj startUpInIslandReader: self.
	next == lastObj] whileFalse:[obj := next].
	^count"! !

!IslandReader methodsFor: 'private' stamp: 'sn 10/24/2018 21:50'!
startUpFromOld: firstObj to: lastObj
	"Send startUp messages to all objects"
	| obj next count |
	count := 0.
	obj := firstObj.
	"Note: This order has been carefully chosen so that even if the object #becomes: itself into something different it will work. This is needed for example by MethodDictionary."
	[
	self halt.next := obj nextObject.
	count := count + 1.
	(obj class = MethodDictionary) ifTrue:[self halt. Transcript show: obj; cr].
	obj startUpInIslandReader: self.
	next == lastObj] whileFalse:[obj := next].
	^count! !

!IslandReader methodsFor: 'reading' stamp: 'ar 6/17/2005 18:59'!
swapInObjectFrom: aStream
	| roots data |
	data := ByteArray new: aStream size.
	aStream next: data size into: data startingAt: 1.
	dataStream := IslandStream with: data.
	dataStream reset.
	(dataStream nextUUID = self class objectVersionID) 
		ifFalse:[^self error: 'Not a croquet file'].
	imports := self importNext.
	segment := self importNext.
	roots := self loadSegmentFrom: segment outPointers: imports.
	^roots at: 1! !

Object subclass: #IslandWriter
	instanceVariableNames: 'archive roots dataStream splObjects segment imports unblessed'
	classVariableNames: 'ArchiveVersionID CodeArray CodeByteString CodeByteSymbol CodeCharacter CodeClass CodeColor CodeConstructor CodeFalse CodeFieldDefinition CodeFloat CodeFont CodeGlobalRef CodeImageSegment CodeLargeNegativeInteger CodeLargePositiveInteger CodeMetaclass CodeMethodTrigger CodeNil CodePoint CodeSmallInteger CodeSpecial CodeTForm CodeTrue CodeUUID CodeWideString CodeWideSymbol ObjectVersionID SignalFailure'
	poolDictionaries: ''
	category: 'Islands-Files'!

!IslandWriter class methodsFor: 'class initialization' stamp: 'ar 6/17/2005 18:58'!
archiveVersionID
	^self subclassResponsibility! !

!IslandWriter class methodsFor: 'class initialization' stamp: 'NiccoKunzmann 7/8/2014 16:06'!
initialize
	"IslandReader initialize"
	"IslandWriter initialize"

	self initializePreferences.

	CodeNil			:= 1.
	CodeTrue 		:= 2.
	CodeFalse 		:= 3.
	CodeSpecial 		:= 4.

	CodeSmallInteger 			:= 5.
	CodeLargePositiveInteger 	:= 6.
	CodeLargeNegativeInteger 	:= 7.
	CodeFloat 					:= 8.
	CodePoint 					:= 9.

	CodeCharacter 	:= 10.
	CodeByteString 	:= 11.
	CodeByteSymbol	:= 12.
	CodeWideString 	:= 13.
	CodeWideSymbol	:= 14.

	CodeClass 		:= 15.
	CodeMetaclass 	:= 16.

	CodeColor 		:= 17.
	CodeFont 		:= 18.
	CodeArray 		:= 19.

	CodeImageSegment 	:= 20.
	CodeMethodTrigger 	:= 21.
	CodeFieldDefinition	:= 22.

	CodeUUID := 23.
	CodeTForm := 24.
	CodeGlobalRef := 25.

	CodeConstructor := 26.

	SignalFailure := true. "by default signal failures"! !

!IslandWriter class methodsFor: 'class initialization' stamp: 'ar 6/16/2005 19:31'!
initializePreferences
	Preferences addPreference: #projectSpaceAnalysis categories: #('Tweak') default: false balloonHelp: 'When true, display a space analysis of the project.'! !

!IslandWriter class methodsFor: 'instance creation' stamp: 'ar 6/16/2005 19:31'!
new
	^self basicNew initialize! !

!IslandWriter class methodsFor: 'class initialization' stamp: 'ar 6/17/2005 18:58'!
objectVersionID
	^self subclassResponsibility! !

!IslandWriter class methodsFor: 'class initialization' stamp: 'ar 6/16/2005 19:31'!
signalFailure
	^SignalFailure! !

!IslandWriter class methodsFor: 'class initialization' stamp: 'ar 6/16/2005 19:31'!
signalFailure: aBool
	SignalFailure := aBool! !

!IslandWriter methodsFor: 'accessing' stamp: 'ar 6/16/2005 19:31'!
archive
	^archive! !

!IslandWriter methodsFor: 'private' stamp: 'ar 6/17/2005 17:43'!
doSpaceAnalysis
	"Capture statistics about the IS and print the number of instances per class and space usage"
	| index sz word hdrBits cc instCount instSpace |
	instCount := IdentityDictionary new.
	instSpace := IdentityDictionary new.
	index := 2. 	"skip version word, first object"
	"go past extra header words"
	hdrBits := (segment at: index) bitAnd: 3.
	hdrBits = 1 ifTrue: [index := index+1].
	hdrBits = 0 ifTrue: [index := index+2].
	[index > segment size] whileFalse:[
		hdrBits := (word := segment at: index) bitAnd: 3.
		hdrBits = 2 ifTrue:[sz := word bitAnd: 16rFFFFFFFC].
		hdrBits = 0 ifTrue:[sz := ((segment at: index-2) bitAnd: 16rFFFFFFFC) + 8].
		hdrBits = 1 ifTrue:[sz := (word bitAnd: "SizeMask" 252) + 4].
		hdrBits = 3 ifTrue:[sz := word bitAnd: "SizeMask" 252].

		"Figure out class name"
		hdrBits = 2 ifTrue:[
			cc := #freeChunk.
		] ifFalse:[ | ccIndex |

			((word := segment at: index) bitAnd: 3) = 2 
				ifTrue: [ccIndex := 0]
				ifFalse:[ccIndex := (word >> 12) bitAnd: 16r1F]."Compact Class field of header word"

			ccIndex = 0 ifTrue:[
				ccIndex := segment at: index-1.
				(ccIndex bitAnd: 16r80000000) = 0 ifTrue:[
					cc := #UserObject.
				] ifFalse:[
					ccIndex := (ccIndex bitAnd: 16r7FFFFFFF) bitShift: -2.
					cc := (imports at: ccIndex) name.
				].
			] ifFalse:[
				cc := (Smalltalk compactClassesArray at: ccIndex) name.
			].
		].
false ifTrue:[
	Transcript cr; nextPutAll: cc; space; print: sz; flush.
].
		instCount at: cc put: (instCount at: cc ifAbsent:[0]) + 1.
		instSpace at: cc put: (instSpace at: cc ifAbsent:[0]) + sz.

		"objectAfter:"
		sz := ((word := segment at: index "header") bitAnd: 3) = 2   "free block?"
			ifTrue: [word bitAnd: 16rFFFFFFFC]
			ifFalse: [(word bitAnd: 3) = 0 "HeaderTypeSizeAndClass"
				ifTrue: [(segment at: index-2) bitAnd: 16rFFFFFFFC]
				ifFalse: [word bitAnd: "SizeMask" 252]].
		index := index + (sz>>2).
		"adjust past extra header words"
		(hdrBits := (segment atPin: index) bitAnd: 3) = 3 ifFalse:[
			"If at end, header word will be garbage.  This is OK"
			hdrBits = 1 ifTrue: [index := index+1].
			hdrBits = 0 ifTrue: [index := index+2].
		].
	].
	^{instCount. instSpace}! !

!IslandWriter methodsFor: 'export' stamp: 'ar 6/16/2005 19:38'!
export: obj
	(splObjects includesKey: obj)
		ifTrue:[self exportSpecial: obj]
		ifFalse:[obj exportOnIslandWriter: self].
! !

!IslandWriter methodsFor: 'export' stamp: 'NiccoKunzmann 7/8/2014 16:05'!
export: aConstructor as: anObject
	"export: aClass as: arrayOfAllAttributes where the class has an constructor method
		aClass>>importFromIslandReader: anObject "
	dataStream nextPut: CodeConstructor.
	self export: aConstructor.
	self export: anObject.! !

!IslandWriter methodsFor: 'export' stamp: 'ar 6/16/2005 19:31'!
exportArray: array
	"Export an array"
	dataStream nextPut: CodeArray.
	dataStream nextUInt32Put: array size.
	array do:[:each| self export: each].! !

!IslandWriter methodsFor: 'export' stamp: 'ar 6/16/2005 19:31'!
exportBoolean: aBool
	dataStream nextPut: (aBool ifTrue:[CodeTrue] ifFalse:[CodeFalse]).! !

!IslandWriter methodsFor: 'export' stamp: 'ar 6/16/2005 19:31'!
exportCharacter: aCharacter
	dataStream nextPut: CodeCharacter.
	dataStream nextUInt32Put: aCharacter asInteger.! !

!IslandWriter methodsFor: 'export' stamp: 'ar 6/16/2005 19:32'!
exportClass: aClass
	"Export class/version"
	(IslandReader isBlessedVersionID: aClass classVersionID) ifFalse:[
		self flag: #fixMe. "Should we raise an exception instead?"
		unblessed add: aClass.
	].
	dataStream nextPut: CodeClass.
	self export: aClass name.
	self export: aClass format.
	self export: aClass allInstVarNames.
	self export: aClass classVersionID.
! !

!IslandWriter methodsFor: 'export' stamp: 'ar 6/16/2005 19:31'!
exportColor: aColor
	"Export a color"
	dataStream nextPut: CodeColor.
	dataStream nextUInt32Put: aColor pixelValue32.! !

!IslandWriter methodsFor: 'export' stamp: 'ar 6/17/2005 18:54'!
exportFarRef: farRef
	^self exportProhibited: farRef! !

!IslandWriter methodsFor: 'export' stamp: 'ar 6/16/2005 19:31'!
exportFieldDefinition: fieldDef
	dataStream nextPut: CodeFieldDefinition.
	self export: fieldDef home name.
	self export: fieldDef name.! !

!IslandWriter methodsFor: 'export' stamp: 'ar 6/16/2005 19:31'!
exportFloat: aFloat
	"Export a literal"
	dataStream nextPut: CodeFloat.
	dataStream nextUInt32Put: (aFloat basicAt: 1).
	dataStream nextUInt32Put: (aFloat basicAt: 2).! !

!IslandWriter methodsFor: 'export' stamp: 'ar 6/16/2005 19:31'!
exportFont: aFont
	"Export a font"
	dataStream nextPut: CodeFont.
	self export: aFont familyName.
	self export: aFont height.
	self export: aFont emphasis.
! !

!IslandWriter methodsFor: 'export' stamp: 'ar 6/16/2005 19:31'!
exportImageSegment: data
	dataStream nextPut: CodeImageSegment.
	dataStream nextUInt32Put: data size.
	dataStream nextWordsPutAll: data.
! !

!IslandWriter methodsFor: 'export' stamp: 'ar 6/16/2005 19:31'!
exportInteger: anInteger
	"Export an integer"
	(anInteger class == SmallInteger) ifTrue:[
		dataStream nextPut: CodeSmallInteger.
		^dataStream nextInt32Put: anInteger.
	].
	anInteger positive ifTrue:[
		dataStream nextPut: CodeLargePositiveInteger.
		dataStream nextUInt32Put: anInteger size.
		dataStream nextPutAll: anInteger asByteArray.
	] ifFalse:[
		dataStream nextPut: CodeLargeNegativeInteger.
		dataStream nextUInt32Put: anInteger size.
		dataStream nextPutAll: anInteger asByteArray.
	].
! !

!IslandWriter methodsFor: 'export' stamp: 'ar 6/16/2005 19:31'!
exportMetaclass: aClass
	"Export class/version"
	dataStream nextPut: CodeMetaclass.
	self exportClass: aClass theNonMetaClass.! !

!IslandWriter methodsFor: 'export' stamp: 'ar 6/16/2005 19:31'!
exportMethodTrigger: aTrigger
	dataStream nextPut: CodeMethodTrigger.
	self export: aTrigger home name.
	self export: aTrigger selector.! !

!IslandWriter methodsFor: 'export' stamp: 'ar 6/16/2005 19:31'!
exportNil: undefinedObj
	"Export a nil"
	dataStream nextPut: CodeNil.! !

!IslandWriter methodsFor: 'export' stamp: 'ar 6/16/2005 19:31'!
exportPoint: aPoint
	dataStream nextPut: CodePoint.
	self export: aPoint x.
	self export: aPoint y.! !

!IslandWriter methodsFor: 'export' stamp: 'ar 6/17/2005 17:52'!
exportProcess: aProcess
	^self exportProhibited: aProcess! !

!IslandWriter methodsFor: 'export' stamp: 'ar 6/17/2005 17:52'!
exportProhibited: anObject
	self error: anObject class name,': Export problem (proceed to ignore)'.
	^self export: nil! !

!IslandWriter methodsFor: 'export' stamp: 'ar 6/16/2005 19:31'!
exportSpecial: special
	dataStream nextPut: CodeSpecial.
	self export: (splObjects at: special).! !

!IslandWriter methodsFor: 'export' stamp: 'ar 6/16/2005 19:47'!
exportString: aString
	aString class isBytes ifTrue:[
		dataStream nextPut: CodeByteString.
		dataStream nextUInt32Put: aString size.
		dataStream nextPutAll: aString asByteArray.
	] ifFalse:[
		dataStream nextPut: CodeWideString.
		dataStream nextUInt32Put: aString size.
		dataStream nextWordsPutAll: aString.
	].! !

!IslandWriter methodsFor: 'export' stamp: 'ar 6/16/2005 19:47'!
exportSymbol: aSymbol
	aSymbol class isBytes ifTrue:[
		dataStream nextPut: CodeByteSymbol.
		dataStream nextUInt32Put: aSymbol size.
		dataStream nextPutAll: aSymbol asByteArray.
	] ifFalse:[
		dataStream nextPut: CodeWideSymbol.
		dataStream nextUInt32Put: aSymbol size.
		dataStream nextWordsPutAll: aSymbol.
	].! !

!IslandWriter methodsFor: 'export' stamp: 'ar 6/17/2005 18:55'!
exportTForm: aTForm
	"Export a TForm"
	^self exportProhibited: aTForm.! !

!IslandWriter methodsFor: 'export' stamp: 'ar 6/16/2005 19:31'!
exportUUID: aUUID
	dataStream nextPut: CodeUUID.
	dataStream nextPutAll: aUUID.! !

!IslandWriter methodsFor: 'export' stamp: 'ar 6/17/2005 17:53'!
exportUnknown: anObject
	^self exportProhibited: anObject! !

!IslandWriter methodsFor: 'public' stamp: 'ar 6/16/2005 19:31'!
failed
	^segment == nil! !

!IslandWriter methodsFor: 'private' stamp: 'ar 12/11/2005 00:02'!
filterArray: array using: wellKnown
	^array allSatisfy:[:each|
		(IslandArgumentCopier isPassByIdentity: each)
			or:[(wellKnown includes: each class)
			or:[each class == Array and:[self filterArray: each using: wellKnown]]]].! !

!IslandWriter methodsFor: 'private' stamp: 'ar 12/11/2005 00:04'!
filterOut: out
	| wellKnown aClass |
	wellKnown := IdentitySet new.
	#(
		"Special references from the islands framework"
		FarRef TFarRef
		"Literals shared via compiled methods"
		String Symbol "for < 3.8"
		ByteString WideString ByteSymbol WideSymbol " >= 3.8"
		"Other objects we know how to deal with"
		Color TranslucentColor Point
		"CompiledMethod since contexts from scripts may refer to them"
		CompiledMethod
		"All font classes"
		AbstractFont LinedTTCFont TTCFontSet StrikeFontSet StrikeFont FixedFaceFont MultiTTCFont TTCFont
	) do:[:clsName|
		aClass := Smalltalk classNamed: clsName.
		aClass ifNotNil:[wellKnown add: aClass].
	].

	^out reject:[:each|
		(IslandArgumentCopier isPassByIdentity: each)
			or:[(splObjects includesKey: each)
			or:[(wellKnown includes: each class)
			or:[each class == Array and:[self filterArray: each using: wellKnown]]]]].
! !

!IslandWriter methodsFor: 'accessing' stamp: 'ar 6/16/2005 19:31'!
imports
	^imports! !

!IslandWriter methodsFor: 'initialize' stamp: 'ar 1/24/2006 17:59'!
initialize
	super initialize.
	unblessed := IdentitySet new.
	archive := ZipArchive new.
	splObjects := IdentityDictionary new.
	splObjects at: TextStyle default put: #TextStyle::Default.
	TextStyle actualTextStyles keysAndValuesDo:[:key :val|
		splObjects at: val put: ('TextStyle::', key) asSymbol.
	].
	Cursor classPool keysAndValuesDo:[:key :val|
		(val isKindOf: Cursor) ifTrue:[
			splObjects at: val put: ('Cursor::', key) asSymbol.
		].
	].
	Smalltalk at: #CCursorLibrary ifPresent:[:aClass|
		aClass default keysAndValuesDo:[:key :val|
			splObjects at: val put: ('CCursorLibrary::', key) asSymbol.
		].
	].
	Smalltalk at: #CIconLibrary ifPresent:[:aClass|
		aClass default icons keysAndValuesDo:[:key :val|
			splObjects at: val put: ('CIconLibrary::Default::', key) asSymbol.
		].
	].
	Smalltalk at: #MenuIcons ifPresent:[:mic|
		(mic classPool at: #Icons) keysAndValuesDo:[:key :val|
			splObjects at: val put: ('MenuIcons::', key) asSymbol.
		].
	].
	Smalltalk at: #ScriptingSystem ifPresent:[:scs|
		scs formDictionary keysAndValuesDo:[:key :val|
			splObjects at: val put: ('ScriptingSystem::Forms::', key) asSymbol.
		].
	].
	Smalltalk at: #CIdentityTransform ifPresent:[:aClass|
		splObjects 
			at: (aClass classPool at: #Default) 
			put: #'CIdentityTransform::Default'.
	].
	Smalltalk at: #IdentityTransform ifPresent:[:aClass|
		splObjects 
			at: (aClass classPool at: #Default) 
			put: #'IdentityTransform::Default'.
	].! !

!IslandWriter methodsFor: 'private' stamp: 'ar 9/15/2005 21:30'!
lockRefsDuring: aBlock
	"Probably want to do this better some time down the road but we need to lock 
	(e.g., strongly hold onto) all far refs in order to force them being outPointers.
	If we don't, then a far ref might be traced by the IS mechanism below and we
	find the value of the far ref in the outPointers (or not even anything!!).
	LATER: Consider whether IS mechanism is really what we want here. Might be
	much easier to markAndTrace island explicitly given the known boundaries of
	#passByIdentity:, method literal, and far ref. Might even be faster..."
	| farRefLock symbolLock |
	farRefLock := FarRef allInstances.
	farRefLock. "fake usage"

	"Same for symbols"
	symbolLock := Symbol allSymbols.
	symbolLock. "fake usage"

	^aBlock value.! !

!IslandWriter methodsFor: 'private' stamp: 'ar 11/8/2005 18:36'!
openWorkspace: contents label: label
	(Workspace new contents: contents) openLabel: label! !

!IslandWriter methodsFor: 'private' stamp: 'ar 6/16/2005 19:31'!
printOutPointerAnalysis: outPointers on: aStream
	"Capture statistics about the IS and print the number of instances per class and space usage"
	| bag sorted |
	aStream nextPutAll:'Far refs:'.
	aStream cr; nextPutAll:'-----------'.
	outPointers do:[:each| each class == FarRef ifTrue:[
		aStream cr; nextPutAll: (each send: #class) name; space; print: each.
	]].

	aStream cr; nextPutAll:'Out pointers:'.
	aStream cr; nextPutAll:'---------------'.

	bag := Bag new.
	outPointers do:[:each| bag add: each class name].
	sorted := bag sortedElements asArray sort:[:a1 :a2|
		a1 value = a2 value ifTrue:[
			(a1 key endsWith: ' class') = (a2 key endsWith: ' class')
				ifTrue:[a1 key <= a2 key]
				ifFalse:[a2 key endsWith: ' class'].
		] ifFalse:[a1 value > a2 value]].
	sorted do:[:assoc|
		aStream cr; nextPutAll: assoc key; tab.
		aStream print: assoc value; nextPutAll: ' instances '.
	].
! !

!IslandWriter methodsFor: 'private' stamp: 'ar 6/16/2005 19:31'!
printSpaceAnalysisOn: aStream
	"Capture statistics about the IS and print the number of instances per class and space usage"
	| instCount instSpace sorted sum1 sum2 data |
	aStream cr; nextPutAll:'Space analysis:'.
	aStream cr; nextPutAll:'------------------'.
	data := self doSpaceAnalysis.
	instSpace := data last.
	instCount := data first.
	sorted := SortedCollection sortBlock:[:a1 :a2| a1 value >= a2 value].
	instSpace associationsDo:[:a| sorted add: a].
	sorted do:[:assoc|
		aStream cr; nextPutAll: assoc key; tab.
		aStream print: (instCount at: assoc key); nextPutAll:' instances '.
		aStream print: assoc value; nextPutAll: ' bytes '.
	].
	sum1 := instCount inject: 0 into:[:sum :n| sum + n].
	sum2 := instSpace inject: 0 into:[:sum :n| sum + n].
	aStream cr; cr.
	aStream print: sum1; nextPutAll:' instances '.
	aStream print: sum2; nextPutAll: ' bytes '.
! !

!IslandWriter methodsFor: 'private' stamp: 'ar 6/16/2005 19:37'!
rootsOf: anIsland
	^anIsland roots! !

!IslandWriter methodsFor: 'public' stamp: 'sn 2/22/2018 22:32'!
saveObjectOn: file
	"This is to be called after 'snapshot', and only once per snapshot.'"

	| data |
	dataStream := IslandStream on: (ByteArray new: segment size * 5).
	dataStream nextPutAll: self class objectVersionID.

	"Export data on dataStream"

	self export: imports.
	self exportImageSegment: segment.

	data := dataStream contents.

	segment := nil.	 "Make these unreachable before the zip phase; every little bit helps."
	imports := nil.
	dataStream := nil.

	archive 
		addDeflateString: data
		as: 'object.isg'.
	archive writeTo: file prepending: self class archiveVersionID.
	^data! !

!IslandWriter methodsFor: 'accessing' stamp: 'ar 6/16/2005 19:31'!
segment
	^segment! !

!IslandWriter methodsFor: 'public' stamp: 'ar 12/8/2009 20:13'!
snapshot: island
	| segSize remain lastObject |
	lastObject := Object new.

	"assemble the roots for this island"
	roots := self rootsOf: island.
"	QLogger logInfo: 'SNAPSHOT at island time ', ((island send:[ : o | o time ])asInteger).	"
	self validateRoots: roots.

	self lockRefsDuring:[
		"Create image segment"
		segSize := 1024*1024.
		[
		segment := nil.		"Make these unreachable, before reallocating."
		imports := nil.
		segment := WordArrayForSegment new: segSize.
		imports := Array new: segSize // 4.
		(self
			storeSegmentFor: roots
			into: segment
			outPointers: imports) == nil
		] whileTrue:[segSize := (segSize * 2)].
	].
	remain := self filterOut: imports.
	remain size > 0 ifTrue:[
		| tracer info |
		SignalFailure == true ifTrue:[CheckpointFailure signal].
		tracer := SnapshotTracer new.
		tracer lastObject: lastObject.
		tracer roots: roots.
		tracer goals: remain.
		tracer segment: segment.
		tracer outPointers: imports.
		tracer context: thisContext sender.
		info := tracer search.
		self openWorkspace: info label: 'WARNING!!!!!!'.
		tracer basicInspect.
		^nil
	].
	Preferences projectSpaceAnalysis ifTrue:[
		| info |
		info := String streamContents:[:s|
			self printOutPointerAnalysis: imports on: s.
			s cr; cr.
			self printSpaceAnalysisOn: s.
		].
		self openWorkspace: info label: 'Info: ', island name.
	].! !

!IslandWriter methodsFor: 'private' stamp: 'sn 2/22/2018 22:09'!
storeSegmentFor: rootsArray into: segmentWordArray outPointers: outPointerArray
	"This primitive will store a binary image segment (in the same format as the Squeak image file) of the receiver and every object in its proper tree of subParts (ie, that is not refered to from anywhere else outside the tree).  Note: all elements of the reciever are treated as roots indetermining the extent of the tree.  All pointers from within the tree to objects outside the tree will be copied into the array of outpointers.  In their place in the image segment will be an oop equal to the offset in the outpointer array (the first would be 4). but with the high bit set."

	"The primitive expects the array and wordArray to be more than adequately long.  In this case it returns normally, and truncates the two arrays to exactly the right size.  If either array is too small, the primitive will fail, but in no other case."

	<primitive: 98 error: ec>	"successful completion returns self"
	^nil							"failure returns nil"! !

!IslandWriter methodsFor: 'public' stamp: 'ar 6/17/2005 18:59'!
swapOutObjectOn: aStream

	self flag: #fixMe.	"We ought to be able to use the file"
	dataStream := IslandStream on: (ByteArray new: segment size * 5).
	dataStream nextPutAll: self class objectVersionID.

	"Export data on dataStream"
	self export: imports.
	self exportImageSegment: segment.
	aStream nextPutAll: dataStream contents.
! !

!IslandWriter methodsFor: 'private' stamp: 'ar 6/18/2005 14:11'!
traceFind: rogueObject
	"Use a checkpoint tracer to find references to the rogue object"
	| tracer info |
	tracer := SnapshotTracer new.
	tracer goals: {rogueObject}.
	tracer segment: segment.
	tracer outPointers: imports.
	info := tracer searchInSegment.
	info inspect.
	tracer basicInspect.! !

!IslandWriter methodsFor: 'accessing' stamp: 'ar 6/16/2005 19:31'!
unblessed
	^unblessed! !

!IslandWriter methodsFor: 'private' stamp: 'ar 6/23/2005 16:35'!
validateRoots: roots
	"Validate the roots to be written out. This is a good thing to do since some objects are unique (nil, true, false) but having them as roots for an island would *force* these to be duplicated which causes major problems down the road. Subclasses may override this method to do more or less extensive checks."
	roots do:[:obj|
		(IslandArgumentCopier isPassByIdentity: obj)
			ifTrue:[self error:'Cannot have passByIdentity: roots'].
	].
! !

!Object class methodsFor: '*Islands' stamp: 'ar 1/17/2005 21:19'!
howToPassAsArgument
	"By default, pass all objects by far proxy"
	^#passByProxy:! !

!Object methodsFor: '*Islands-Support-description' stamp: ''!
checkpointDescription
	^ self class name! !

!Object methodsFor: '*Islands-Support-description' stamp: 'jrd 12/21/2009 15:42'!
collectReplicateValues
	"(self future new: Object) future collectReplicateValues"

	| replicateValues replicateID |
	replicateValues := OrderedCollection new.
	replicateID := TObjectID for: self.
	self island register: replicateValues name: replicateID.
	self island signal: #collectReplicateValues withArguments: { self . replicateValues . replicateID copy}.
	^replicateValues! !

!Object methodsFor: '*Islands' stamp: 'ar 6/16/2005 19:42'!
exportOnIslandWriter: writer
	^writer exportUnknown: self! !

!Object methodsFor: '*Islands' stamp: 'ar 1/14/2005 13:15'!
isFarRef
	"Am I a reference to an object on another island?"
	^false! !

!Object methodsFor: '*Islands' stamp: 'ar 5/22/2005 15:38'!
island
	"Answer the receiver's island"
	^Processor activeIsland! !

!Object methodsFor: '*Islands' stamp: 'ar 5/22/2005 13:47'!
myFarRef
	^nil! !

!Object methodsFor: '*Islands' stamp: 'ar 5/22/2005 13:47'!
myFarRef: farRef
	^nil! !

!Object methodsFor: '*Islands' stamp: 'ar 6/16/2005 19:57'!
startUpInIslandReader: reader
	"The receiver has been loaded from a binary file."! !

!Object methodsFor: '*Islands' stamp: 'ar 5/22/2005 15:34'!
valueOn: island
	"Answer the receiver's value on the given island."
	^self! !

Object subclass: #SnapshotTracer
	instanceVariableNames: 'roots goals parents context lastObject log segment outPointers objCount'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Islands-Support'!

!SnapshotTracer commentStamp: '<historical>' prior: 0!
CheckpointTracer is a tool for debugging problems with Island checkpointing. It works in precisely the same way that the image segment creation primitive works but is capable of finding the reference paths to any of the goal objects (which are the ones that were identified as problematic "outside objects"). This information can in turn be used to identify what has gone wrong.!

!SnapshotTracer methodsFor: 'searching' stamp: 'ar 5/16/2005 11:53'!
buildRawList: goal
	| list obj |
	list := OrderedCollection new.
	obj := goal.
	[list addFirst: obj.
	obj := parents at: obj ifAbsent:[nil].
	obj == nil] whileFalse.
	^list! !

!SnapshotTracer methodsFor: 'accessing' stamp: 'ar 5/15/2005 00:40'!
context
	^context! !

!SnapshotTracer methodsFor: 'accessing' stamp: 'ar 5/15/2005 00:41'!
context: aContext
	context := aContext! !

!SnapshotTracer methodsFor: 'searching' stamp: 'ar 5/21/2005 11:13'!
countObjects
	| object count |
	count := 0.
	object := self someObject.
	[lastObject == object] whileFalse: [
		count := count + 1.
		object := object nextObject.
	].
	^objCount := count! !

!SnapshotTracer methodsFor: 'searching' stamp: 'ar 5/21/2005 11:34'!
countSegment
	| index hdrBits count |
	count := 0.
	index := 2.
	"go past extra header words"
	(hdrBits := (segment atPin: index) bitAnd: 3) = 1 ifTrue: [index := index+1].
	hdrBits = 0 ifTrue: [index := index+2].
	[count := count + 1.
	index := self objectAfter: index.
	index > segment size] whileFalse.
	^objCount := count! !

!SnapshotTracer methodsFor: 'searching' stamp: 'ar 5/24/2005 14:59'!
gcTrace: todo label: aString
	"Search for a reference path to the goals"
	| obj max step next now |
	aString
		displayProgressAt: Sensor cursorPoint 
		from: 0 to: 200 during:[:bar|

		bar value: 0.
		max := objCount.
		step := (max // 200) max: 1.
		next := step.
		now := 0.
		[todo isEmpty] whileFalse:[
			(now := now + 1) > next ifTrue:[
				bar value: next // step.
				next := next + step.
			].
			obj := todo removeLast.
			self markAndTrace: obj with: todo.
		].
	].! !

!SnapshotTracer methodsFor: 'accessing' stamp: 'ar 5/14/2005 19:54'!
goals
	^goals! !

!SnapshotTracer methodsFor: 'accessing' stamp: 'ar 5/14/2005 19:55'!
goals: aCollection
	goals := IdentitySet withAll: aCollection.! !

!SnapshotTracer methodsFor: 'searching' stamp: 'ar 5/16/2005 15:32'!
isMarked: child
	"Answer whether child has been marked already"
	child ifNil:[^true].
	child class == SmallInteger ifTrue:[^true].
	^parents includesKey: child! !

!SnapshotTracer methodsFor: 'accessing' stamp: 'ar 5/15/2005 18:31'!
lastObject
	^lastObject! !

!SnapshotTracer methodsFor: 'accessing' stamp: 'ar 5/15/2005 18:31'!
lastObject: obj
	lastObject := obj! !

!SnapshotTracer methodsFor: 'searching' stamp: 'jcg 5/28/2006 22:08'!
markAndTrace: obj with: aList
	| objClass child |
	objClass := obj class.

	"Trace class"
	child := objClass.
	(self isMarked: child) ifFalse:[
		parents at: child put: obj.
		aList add: child].

	objClass == SmallInteger ifTrue:[^false].

	objClass == CompiledMethod ifTrue:[
		"Special case for CompiledMethod"
		1 to: obj numLiterals do:[:i|
			child := obj literalAt: i.
			(self isMarked: child) ifFalse:[
				parents at: child put: obj.
				aList add: child]].
		^false].

	objClass isPointers ifFalse:[^false].

	1 to: objClass instSize do:[:i|
		child := obj instVarAt: i.
		(self isMarked: child) ifFalse:[
			parents at: child put: obj.
			aList add: child]].
	objClass isVariable ifFalse:[^false].
	objClass isWeak ifTrue:[^false].
	obj == outPointers ifTrue: [^false].
	1 to: obj basicSize do:[:i|
		child := obj basicAt: i.
		(self isMarked: child) ifFalse:[
			parents at: child put: obj.
			aList add: child]].
	^false! !

!SnapshotTracer methodsFor: 'searching' stamp: 'ar 5/21/2005 11:34'!
objectAfter: ind
	"Return the object or free chunk immediately following the given object or free chunk in the segment.  *** Warning: When class ObjectMemory change, be sure to change it here. ***"

	| sz word newInd hdrBits |
	sz := ((word := segment at: ind "header") bitAnd: 3) = 2   "free block?"
		ifTrue: [word bitAnd: 16rFFFFFFFC]
		ifFalse: [(word bitAnd: 3) = 0 "HeaderTypeSizeAndClass"
			ifTrue: [(segment at: ind-2) bitAnd: 16rFFFFFFFC]
			ifFalse: [word bitAnd: "SizeMask" 252]].

	newInd := ind + (sz>>2).
	"adjust past extra header words"
	(hdrBits := (segment atPin: newInd) bitAnd: 3) = 3 ifTrue: [^ newInd].
		"If at end, header word will be garbage.  This is OK"
	hdrBits = 1 ifTrue: [^ newInd+1].
	hdrBits = 0 ifTrue: [^ newInd+2].
	^ newInd	"free"! !

!SnapshotTracer methodsFor: 'accessing' stamp: 'ar 5/21/2005 11:11'!
outPointers
	^outPointers! !

!SnapshotTracer methodsFor: 'accessing' stamp: 'ar 5/21/2005 11:11'!
outPointers: out
	outPointers := out.! !

!SnapshotTracer methodsFor: 'searching' stamp: 'mtf 1/20/2010 19:21'!
printPath: goal
	"Print the reference path to the goal we found"
	| list obj parent key |
	(self isMarked: goal) ifFalse:[
		log cr; cr; nextPutAll: (goal name printString truncateTo: 50), ' : ', goal class name.
		log cr; nextPutAll: '<<no reference path could be found>>'.
		^self].
	list := self buildRawList: goal.

	obj := list removeFirst.
	log cr; cr; nextPutAll: 'root: ', (obj == Smalltalk specialObjectsArray ifTrue:['Smalltalk specialObjects'] ifFalse:[obj name]), ' (', obj class name,')'.
	[list isEmpty] whileFalse:[
		parent := obj.
		obj := list removeFirst.
		key := nil.
		(list size >= 3 and:[parent class includesBehavior: CObject]) ifTrue:[
			((obj == (parent instVarNamed: 'myProperties')) and:[
				(list first == (obj instVarNamed: 'array'))]) ifTrue:[
					key := list second key.
					list removeFirst; removeFirst.
					obj := list removeFirst.
				].
		].
		(parent class includesBehavior: Dictionary) ifTrue:[
			list size >= 2 ifTrue:[
				key := parent keyAtValue: list second ifAbsent: [].
				key ifNotNil:[
					obj := list removeFirst; removeFirst.
				].
			].
		].
		key ifNil:[parent class == obj ifTrue: [key := 'CLASS']].
		key ifNil:[
			1 to: parent class instSize do: [:i | 
				key ifNil:[ (parent instVarAt: i) == obj ifTrue:[
						key := parent class instVarNameForIndex: i.
				]]]].
		key ifNil:[parent class isVariable ifTrue:[
			1 to: parent basicSize do: [:i | 
				key ifNil:[(parent basicAt: i) == obj ifTrue: [
					key := i printString.
				]]]]].
		key ifNil:[key := '???'].
		log cr; nextPutAll: key asString, ': ' , obj checkpointDescription
	].! !

!SnapshotTracer methodsFor: 'accessing' stamp: 'ar 5/14/2005 19:54'!
roots
	^roots! !

!SnapshotTracer methodsFor: 'accessing' stamp: 'ar 5/14/2005 19:54'!
roots: aCollection
	roots := aCollection.! !

!SnapshotTracer methodsFor: 'searching' stamp: 'ar 5/21/2005 11:49'!
search
	"Search for a reference path to the goals"
	| todo |
	self countObjects.
	log := WriteStream on: String new.
	parents := IdentityDictionary new: objCount.
	parents at: roots put: nil.
	roots do:[:each| parents at: each put: nil].
	parents at: self put: nil. "avoid tracing me"
	todo := OrderedCollection new.

	"Prepare a number of good roots"
	context ifNotNil:[
		parents at: context put: nil.
		todo add: context.
	].
	Smalltalk allClassesDo:[:cls| 
		parents at: cls put: nil.
		todo add: cls.
	].
	parents at: Smalltalk specialObjectsArray put: nil.
	todo add: Smalltalk specialObjectsArray.

	"GC mark all objects outside"
	self gcTrace: todo label: 'Checkpoint failure - searching image references...'.
	"Print reference paths to the goals"
	log cr; cr; nextPutAll: 'Outside references:'.
	goals do:[:each| self printPath: each].

	"Remove goals from marked objects again so we can find them from the internal references"
	goals do:[:each| parents removeKey: each ifAbsent:[]].

	"Setup next round for gc tracing"
	roots := ImageSegment basicNew
		loadSegmentFrom: segment copy
		outPointers: outPointers.
	roots do:[:each| parents at: each put: nil].
	todo := OrderedCollection new.
	todo addAll: roots.
	self countSegment.
	self gcTrace: todo label: 'Checkpoint failure - searching segment references...'.
	"Print reference paths to the goals"
	log cr; cr; nextPutAll: 'Inside references:'.
	goals do:[:each| self printPath: each].

false ifTrue:[
	"Check for eventual unmarked objects"
	todo := self unmarkedObjects.
	todo inspect.
].
	^log contents! !

!SnapshotTracer methodsFor: 'searching' stamp: 'ar 5/26/2005 22:41'!
searchInSegment
	"Setup next round for gc tracing"
	| todo |
	self countSegment.
	log ifNil:[log := WriteStream on: String new].
	roots := ImageSegment basicNew
		loadSegmentFrom: segment copy
		outPointers: outPointers.
	parents := IdentityDictionary new: objCount.
	roots do:[:each| parents at: each put: nil].
	outPointers do:[:each| parents at: each put: nil].
	todo := OrderedCollection new.
	todo addAll: roots.
	self gcTrace: todo label: 'Checkpoint failure - searching segment references...'.
	"Print reference paths to the goals"
	log cr; cr; nextPutAll: 'Inside references:'.
	goals do:[:each| self printPath: each].
	^log contents! !

!SnapshotTracer methodsFor: 'searching' stamp: 'ar 5/26/2005 22:39'!
searchInSystem
	"Search for a reference path to the goals"
	| todo |
	self countObjects.
	log ifNil:[log := WriteStream on: String new].
	parents := IdentityDictionary new: objCount.
	parents at: roots put: nil.
	roots do:[:each| parents at: each put: nil].
	parents at: self put: nil. "avoid tracing me"
	todo := OrderedCollection new.

	"Prepare a number of good roots"
	context ifNotNil:[
		parents at: context put: nil.
		todo add: context.
	].
	Smalltalk allClassesDo:[:cls| 
		parents at: cls put: nil.
		todo add: cls.
	].
	parents at: Smalltalk specialObjectsArray put: nil.
	todo add: Smalltalk specialObjectsArray.

	"GC mark all objects outside"
	self gcTrace: todo label: 'Checkpoint failure - searching image references...'.
	"Print reference paths to the goals"
	log cr; cr; nextPutAll: 'Outside references:'.
	goals do:[:each| self printPath: each].

	^log contents! !

!SnapshotTracer methodsFor: 'accessing' stamp: 'ar 5/21/2005 11:10'!
segment
	^segment! !

!SnapshotTracer methodsFor: 'accessing' stamp: 'ar 5/21/2005 11:10'!
segment: seg
	segment := seg! !

!SnapshotTracer methodsFor: 'searching' stamp: 'ar 5/16/2005 15:03'!
unmarkedObjects
	"All objects should be marked by now. If there are any unmarked ones, report them"
	| set obj |
	set := IdentitySet new.
	obj := self someObject.
	[obj == lastObject] whileFalse:[
		(self isMarked: obj) ifFalse:[set add: obj].
		obj := obj nextObject.
	].
	^set! !

Object subclass: #TestVariable
	instanceVariableNames: 'value'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Islands-Support'!

!TestVariable commentStamp: 'ar 1/17/2005 16:58' prior: 0!
A simple test variable.!

!TestVariable methodsFor: 'other' stamp: 'ar 1/14/2005 23:10'!
printOn: aStream
	super printOn: aStream.
	aStream nextPutAll:'(', value,')'.! !

!TestVariable methodsFor: 'accessing' stamp: 'ar 1/14/2005 21:55'!
value
	^value! !

!TestVariable methodsFor: 'accessing' stamp: 'ar 1/14/2005 21:55'!
value: newValue
	value := newValue! !

!CompiledMethod class methodsFor: '*Islands' stamp: 'ar 1/17/2005 21:18'!
howToPassAsArgument
	"Do not copy CompiledMethods"
	^#passByProxy:! !

!Metaclass class methodsFor: '*Islands' stamp: 'ar 5/22/2005 15:30'!
howToPassAsArgument
	"To override implementation for Object - Metaclasses are passByIdentity: too"
	^#passByIdentity:! !

!Metaclass methodsFor: '*Islands' stamp: 'ar 6/16/2005 19:42'!
exportOnIslandWriter: writer
	^writer exportMetaclass: self! !

!ZipArchiveMember methodsFor: '*Islands' stamp: 'ar 6/16/2005 17:56'!
byteStream
	"Answer a suitable byte stream for the contents of the receiver.
	Intended to avoid *any* temporary decompression and conversion.
	However, as a general fallback we simply use #contentStream."
	^(self contentStream)
		binary; 
		yourself! !

!Array methodsFor: '*Islands' stamp: 'ar 6/16/2005 19:39'!
exportOnIslandWriter: writer
	^writer exportArray: self! !

!StrikeFont class methodsFor: '*Islands' stamp: 'ar 1/16/2005 22:01'!
howToPassAsArgument
	"passByIdentity: for rendering purposes"
	^#passByIdentity:! !

Inspector subclass: #FarRefInspector
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Islands-Support'!

!FarRefInspector commentStamp: 'ar 1/17/2005 16:58' prior: 0!
A small hack on inspector to show the contents of a far object.!

!FarRefInspector methodsFor: 'accessing' stamp: 'ar 5/22/2005 13:38'!
actualClassOf: obj
	^obj send: #class to: obj! !

!FarRefInspector methodsFor: 'accessing' stamp: 'ar 5/22/2005 13:38'!
baseFieldList
	"Answer an Array consisting of 'self'
	and the instance variable names of the inspected object."
	(object isFarRef)
		 ifFalse:[^super baseFieldList].
	^(Array with: 'self' with: 'all inst vars')
			, (self actualClassOf: object) allInstVarNames! !

!FarRefInspector methodsFor: 'accessing' stamp: 'ar 1/17/2005 23:01'!
inspectSelection
	"Create and schedule an Inspector on the receiver's model's currently selected object."

	self selectionIndex = 0 ifTrue: [^ self changed: #flash].
	^FarRefInspector openOn: self selection withEvalPane: true.! !

!FarRefInspector methodsFor: 'accessing' stamp: 'ar 5/22/2005 13:39'!
selection
	"The receiver has a list of variables of its inspected object.
	One of these is selected. Answer the value of the selected variable."
	| basicIndex aClass |
	selectionIndex = 0 ifTrue: [^ ''].
	selectionIndex = 1 ifTrue: [^ object].
	selectionIndex = 2 ifTrue: [^ object longPrintString].
	aClass := self actualClassOf: object.
	(selectionIndex - 2) <= aClass instSize
		ifTrue: [^ object farInstVarAt: selectionIndex - 2].
	basicIndex := selectionIndex - 2 - aClass instSize.
	(object basicSize <= (self i1 + self i2)  or: [basicIndex <= self i1])
		ifTrue: [^ object basicAt: basicIndex]
		ifFalse: [^ object basicAt: object basicSize - (self i1 + self i2) + basicIndex]! !

!Form methodsFor: '*Islands' stamp: 'ar 6/16/2005 20:11'!
hackRow: bitThing
	"This method provides an initialization so that BitBlt may be used, eg, to 
	copy ByteArrays and other non-pointer objects efficiently.
	The resulting form looks 1 high, 8 deep, and bitThing-size-in-words wide."
	height := 1.
	depth := 8.
	bitThing class isBits ifFalse: [self error: 'bitThing must be a non-pointer object'].
	bitThing class isBytes
		ifTrue: [width := bitThing basicSize]
		ifFalse: [width := bitThing basicSize*4].
	bits := bitThing! !

Time subclass: #IslandTime
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Islands-Time'!

!IslandTime commentStamp: '<historical>' prior: 0!
I represent the system real-time clock. I implement a few methods so that I can be used both, as a time base for Time (e.g., Time maps to per-island  time) as well as independently in expressions such as "RealTimeClock millisecondsToRun:[...]".!

!IslandTime class methodsFor: 'accessing' stamp: 'ar 9/17/2005 23:02'!
millisecondClockValue
	"Answer the number of milliseconds since the millisecond clock was last reset or rolled over."
	^Processor activeIsland millisecondClockValue! !

!IslandTime class methodsFor: 'accessing' stamp: 'ar 9/17/2005 23:02'!
totalSeconds
	"Answer the total seconds since the Squeak epoch: 1 January 1901."
	^Processor activeIsland totalSeconds! !

Time subclass: #RealTime
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Islands-Time'!

!RealTime commentStamp: '<historical>' prior: 0!
I represent the system real-time clock. I implement a few methods so that I can be used both, as a time base for Time (e.g., Time maps to real-time) as well as independently in expressions such as "RealTimeClock millisecondsToRun:[...]".!

!RealTime class methodsFor: 'accessing' stamp: 'eem 9/28/2009 08:19'!
localMicrosecondClock
	"Answer the number of microseconds since the start of the 20th century in local time."
	<primitive: 241>
	^0! !

!RealTime class methodsFor: 'accessing' stamp: 'ar 9/17/2005 18:57'!
millisecondClockValue
	"Answer the number of milliseconds since the millisecond clock was last reset or rolled over.
	Answer 0 if the primitive fails."
	<primitive: 135>
	^0! !

!RealTime class methodsFor: 'accessing' stamp: 'ar 9/17/2005 18:57'!
totalSeconds
	"Answer the total seconds since the Squeak epoch: 1 January 1901.
	Answer 0 if the primitive fails."
	<primitive: 137>
	^0! !

!RealTime class methodsFor: 'accessing' stamp: 'eem 9/28/2009 08:18'!
utcMicrosecondClock
	"Answer the number of microseconds since the start of the 20th century in UTC."
	<primitive: 240>
	^0! !

!Point class methodsFor: '*Islands' stamp: 'ar 1/16/2005 21:59'!
howToPassAsArgument
	"Clone me - might want to change this to passByIdentity: since points are effectively read-only"
	^#passByClone:! !

!Point methodsFor: '*Islands' stamp: 'ar 6/16/2005 19:43'!
exportOnIslandWriter: writer
	^writer exportPoint: self.! !

!ShortIntegerArray methodsFor: '*Islands' stamp: 'sn 2/22/2018 22:23'!
startUpInIslandReader: reader
	"The receiver has been loaded from a binary file."
	(Smalltalk endianness) == (reader endianness)
		ifFalse:[self swapShortObjects].! !

!MethodDictionary methodsFor: '*Islands' stamp: 'sn 2/26/2018 16:53'!
startUpInIslandReader: reader
	"Rehash but without #become:"
	| sz newSelf key |
	sz := self basicSize.
	"newSelf := (MethodDictionary new: sz) init: sz."
	newSelf := (MethodDictionary new: sz) initialize: sz.
	1 to: sz do:[:i | 
		key := self basicAt: i.
		key == nil ifFalse: [newSelf at: key put: (array at: i)].
	].
	1 to: sz do:[:i| self basicAt: i put: (newSelf basicAt: i)].
	array := newSelf array.! !

!Behavior class methodsFor: '*Islands' stamp: 'ar 9/17/2005 19:24'!
howToPassAsArgument
	"passByIdentity: for now - until we have proper behavior representations inside the islands"
	^#passByIdentity:! !

!Behavior methodsFor: '*Islands' stamp: 'ar 1/16/2005 21:55'!
howToPassAsArgument
	"passByIdentity: for now - until we have proper behavior representations inside the islands"
	^#passByIdentity:! !

!ArrayedCollection class methodsFor: '*Islands' stamp: 'ar 1/16/2005 21:54'!
howToPassAsArgument
	"Copy array-like objects by default"
	^#passByCopy:! !

!Float methodsFor: '*Islands' stamp: 'ar 6/16/2005 19:41'!
exportOnIslandWriter: writer
	^writer exportFloat: self! !

!Rectangle class methodsFor: '*Islands' stamp: 'ar 5/22/2005 13:50'!
howToPassAsArgument
	"Copy rectangles. Possibly change to passByIdentity: since rectangles are mostly immutable."
	^#passByClone:! !

!UUID methodsFor: '*Islands' stamp: 'ar 6/16/2005 19:44'!
exportOnIslandWriter: writer
	^writer exportUUID: self! !

!ShortRunArray methodsFor: '*Islands' stamp: 'sn 2/22/2018 22:23'!
startUpInIslandReader: reader
	"The receiver has been loaded from a binary file."
	(Smalltalk endianness) == (reader endianness)
		ifFalse:[self swapRuns].! !

!String methodsFor: '*Islands' stamp: 'ar 6/16/2005 19:44'!
exportOnIslandWriter: writer
	^writer exportString: self! !

!Collection class methodsFor: '*Islands-override' stamp: 'mtf 1/22/2010 17:08'!
randomForPicking
	^ Processor activeIsland random! !

!MethodContext methodsFor: '*Islands-Support-description' stamp: ''!
checkpointDescription
	^ super checkpointDescription , '  ' , self printString asString! !

!AbstractFont methodsFor: '*Islands' stamp: 'ar 6/16/2005 19:38'!
exportOnIslandWriter: writer
	^writer exportFont: self! !

!Semaphore class methodsFor: '*Islands' stamp: 'ar 1/16/2005 22:01'!
howToPassAsArgument
	"passByIdentity: for now; I think we don't need this but some primitives are unhappy with far refs and need some checking before this will work properly"
	^#passByIdentity:! !

!HashedCollection methodsFor: '*Islands' stamp: 'ar 6/16/2005 19:59'!
startUpInIslandReader: reader
	"The receiver has been loaded from a binary file."
	self rehash. "Probably way to conservative but better be safe than sorry"! !

!SoundBuffer methodsFor: '*Islands' stamp: 'sn 2/22/2018 22:23'!
startUpInIslandReader: reader
	"The receiver has been loaded from a binary file."
	(Smalltalk endianness) == (reader endianness)
		ifFalse:[self swapHalves].! !

!Color class methodsFor: '*Islands' stamp: 'ar 1/17/2005 21:31'!
howToPassAsArgument
	"Clone colors"
	^#passByClone:! !

!Color methodsFor: '*Islands' stamp: 'ar 6/16/2005 19:41'!
exportOnIslandWriter: writer
	^writer exportColor: self! !

!Process class methodsFor: '*Islands' stamp: 'ar 6/15/2009 13:38'!
forIsland: anIsland
	"Create the given process explicitly in the context of an island"
	^(self basicNew) setIsland: anIsland; yourself! !

!Process class methodsFor: '*Islands' stamp: 'ar 1/15/2005 23:47'!
howToPassAsArgument
	"Processes are implicitly passByIdentity: - at least as long as we modify them in place (we may change this further down the road)."
	^#passByIdentity:! !

!Process class methodsFor: '*Islands' stamp: 'ar 6/15/2009 13:38'!
new
	^self forIsland: Processor activeIsland! !

!Process methodsFor: '*Islands' stamp: 'mtf 11/2/2011 11:09'!
activateIsland: anIsland during: aBlock
"Set my island to anIsland for the duration of aBlock. Answers the value of aBlock.

If an exception is raised during aBlock, any exception handlers will be evaluated in the context of anIsland, even those below me in the stack. For better exception handling behavior, use #activateIslandWithEscape:during:"

	| priorIsland |
	priorIsland := self island.
	^ [self island: anIsland. aBlock value]
		ensure: [self island: priorIsland]! !

!Process methodsFor: '*Islands' stamp: 'mtf 11/2/2011 10:48'!
activateIslandWithEscape: anIsland during: aBlock
"Set my island to anIsland for the duration of aBlock. Answers the value of aBlock.

If an exception is raised during aBlock, any exception handlers below me in the stack will be evaluated in the old island, and any exception handlers above me will be evaluated in the new island. Due to the recursive call and my re-entrant exception handler, I guarantee this even if exceptions are thrown inside exception handlers"

	| priorIsland |
	priorIsland := self island.
	^ [self activateIsland: anIsland during: aBlock]
		on: Exception do: [:ex | ex resume: (
			self activateIslandWithEscape: priorIsland during: [
				ex rearmHandlerDuring: [ex outer]])].
! !

!Process methodsFor: '*Islands' stamp: 'ar 6/16/2005 19:43'!
exportOnIslandWriter: writer
	^writer exportProcess: self! !

!Process methodsFor: '*Islands' stamp: 'ar 6/15/2009 13:38'!
setIsland: newIsland
	"Initialize processes by default with the inherited island from the parent process"
	island := newIsland.
	island ifNotNil:[island noteNewProcess: self].
! !

!Symbol class methodsFor: '*Islands' stamp: 'ar 1/16/2005 22:01'!
howToPassAsArgument
	"Symbols are immutable"
	^#passByIdentity:! !

!Symbol methodsFor: '*Islands' stamp: 'ar 6/16/2005 19:44'!
exportOnIslandWriter: writer
	^writer exportSymbol: self! !

Association subclass: #IslandVariable
	instanceVariableNames: 'home'
	classVariableNames: 'WriteStatistics'
	poolDictionaries: ''
	category: 'Islands-Kernel'!

!IslandVariable commentStamp: 'ar 1/17/2005 16:55' prior: 0!
I am a variable which can be used with islands for global, class, or pool variables. I manage my value so that the isolation layers between islands stay intact. If the object stored in me is not a passByIdentity: object I store a far reference to the object. When reading my value the far reference gets resolved in the context of the reader's island. Unless used with passByIdentity: objects, island variables are fairly slow.

[**** TBD: We should have an shared variable which ONLY stores passByIdentity: objects and raises an error when anything else is being stored. Then we can actually use that variable directly so reads are zero overhead. IslandVariable would then be the slow default version. ****]

[**** TBD: If we do this, we should have a menu just like for fields and be able to do this on the basis of individual class vars ****]
!

!IslandVariable class methodsFor: 'accessing' stamp: 'ar 1/16/2005 22:02'!
howToPassAsArgument
	"passByIdentity: since variables can be shared and encapsulate contents themselves"
	^#passByIdentity:! !

!IslandVariable class methodsFor: 'accessing' stamp: 'ar 1/16/2005 08:27'!
key: newKey value: newValue
	^self new key: newKey value: newValue! !

!IslandVariable methodsFor: 'accessing' stamp: 'ar 1/16/2005 08:27'!
home
	^home! !

!IslandVariable methodsFor: 'accessing' stamp: 'ar 1/16/2005 08:27'!
home: anObject
	home := anObject.! !

!IslandVariable methodsFor: 'testing' stamp: 'ar 1/16/2005 08:27'!
isSpecialReadBinding
	"Always access via #value"
	^true! !

!IslandVariable methodsFor: 'testing' stamp: 'ar 1/16/2005 08:27'!
isSpecialWriteBinding
	"Always access via #value:"
	^true! !

!IslandVariable methodsFor: 'testing' stamp: 'ar 1/16/2005 08:27'!
isVariableBinding
	"Yes, of course"
	^true! !

!IslandVariable methodsFor: 'accessing' stamp: 'ar 1/16/2005 08:27'!
key: aKey value: anObject 
	"Store the arguments as the variables of the receiver."
	key := aKey.
	self value: anObject.
	^self! !

!IslandVariable methodsFor: 'printing' stamp: 'ar 1/16/2005 08:27'!
printOn: aStream
	"Print as home::key -> value"
	home ifNotNil:[aStream nextPutAll: home name].
	aStream nextPutAll: '::'; nextPutAll: key.
	aStream nextPutAll: ' -> '.
	aStream print: value.! !

!IslandVariable methodsFor: 'accessing' stamp: 'ar 1/17/2005 14:48'!
value
	"Answer the receiver's value. Resolve far refs to the current island."
	| v |
	v := value. "read value atomically"
	v isFarRef 
		ifTrue:[^v varValueOn: Processor activeIsland]
		ifFalse:[^v]
! !

!IslandVariable methodsFor: 'accessing' stamp: 'ar 5/22/2005 14:39'!
value: newValue
	"Assign a new value to this variable. Conceptually, we always store a far ref here but optimize this for #passByIdentity: (clients cannot distinguish this; try it)"
	| v |
	(IslandArgumentCopier isPassByIdentity: newValue)
		ifTrue:[v := newValue]
		ifFalse:[
			"Uncomment this to gather statistics"
			"WriteStatistics ifNotNil:[WriteStatistics add: key]."
			v := newValue island asFarRef: newValue.
		].
	value := v. "Write value atomically"
	^newValue
! !

ProtoObject subclass: #FarRef
	instanceVariableNames: 'myIsland myValue'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Islands-Kernel'!

!FarRef commentStamp: 'ar 1/17/2005 11:50' prior: 0!
I represent a reference to an object living on another island. When I receive a message I pass the message on to the object, ensuring that any arguments are converted properly so that no direct links between islands exists (except those explicitly permitted in the islands framework). Upon return, the returned value is converted similarly.
!

!FarRef class methodsFor: 'accessing' stamp: 'ar 1/17/2005 21:19'!
howToPassAsArgument
	"passByProxy: by definition"
	^#passByProxy:! !

!FarRef class methodsFor: 'recompiling' stamp: 'ar 1/14/2005 22:25'!
newInstanceFrom: oldInstance variable: variable size: instSize map: map
	"Create a new instance of the receiver based on the given old instance.
	The supplied map contains a mapping of the old instVar names into
	the receiver's instVars"
	| new |
	variable
		ifTrue: [new := self basicNew: oldInstance basicSize]
		ifFalse: [new := self basicNew].
	1 to: instSize do: 
		[:offset |  (map at: offset) > 0 ifTrue:
			[new primitiveInstVarAt: offset
					put: (oldInstance primitiveInstVarAt: (map at: offset))]].
	variable 
		ifTrue: [1 to: oldInstance basicSize do: 
					[:offset |
					new primitiveBasicAt: offset put: (oldInstance primitiveBasicAt: offset)]].
	^new! !

!FarRef methodsFor: 'private' stamp: 'ar 1/17/2005 16:35'!
asDestForm
	"We allow this for primitives. 
	TODO: Fix BitBlt so this ain't necessariy (make it passByIdentity:?)"
	^myValue asDestForm! !

!FarRef methodsFor: 'private' stamp: 'ar 1/17/2005 16:35'!
asSourceForm
	"We allow this for primitives. 
	TODO: Fix BitBlt so this ain't necessariy (make it passByIdentity:?)"
	^myValue asSourceForm! !

!FarRef methodsFor: 'copying' stamp: 'tk 6/2/2005 10:18'!
copyToCurrentIsland
	"Return a copy of my value, completely resolved to be in the current Island."

	| copier |
	copier := CDeepCopier new initialize: self initialDeepCopierSize.
	copier forbiddenIsland: myIsland.
	^ self veryDeepCopyWith: copier! !

!FarRef methodsFor: 'debugging' stamp: 'ar 1/13/2005 22:40'!
defaultLabelForInspector
	"Answer the default label to be used for an Inspector window on the receiver."
	self proxyDebuggerMessages ifFalse:[^self class name].
	^'Far:[',(self send: #defaultLabelForInspector),']'! !

!FarRef methodsFor: 'messaging' stamp: 'ar 1/17/2005 11:37'!
doesNotUnderstand: aMessage
	"Send the selector, aSymbol, to the receiver with arguments in argArray."
	^self syncSend: aMessage selector withArguments: aMessage arguments! !

!FarRef methodsFor: 'errors' stamp: 'ar 1/13/2005 22:16'!
error: aString 
	"Throw a generic Error exception."

	^Error new signal: aString! !

!FarRef methodsFor: 'private' stamp: 'ar 6/16/2005 19:41'!
exportOnIslandWriter: writer
	^writer exportFarRef: self! !

!FarRef methodsFor: 'debugging' stamp: 'tk 6/1/2005 17:17'!
farInstVarAt: index 
	"instVarAt: for far references"
	^self send: #instVarAt: with: index! !

!FarRef methodsFor: 'debugging' stamp: 'ar 1/15/2005 14:10'!
farInstVarAt: anInteger put: anObject 
	"Primitive. Store a value into a fixed variable in the receiver. The 
	numbering of the variables corresponds to the named instance variables. 
	Fail if the index is not an Integer or is not the index of a fixed variable. 
	Answer the value stored as the result. Using this message violates the 
	principle that each object has sovereign control over the storing of 
	values into its instance variables. Essential. See Object documentation 
	whatIsAPrimitive."
	^self send: #instVarAt: with: anInteger with: anObject! !

!FarRef methodsFor: 'compatability' stamp: 'mtf 11/2/2011 12:13'!
futureSendAndWait: selector withArguments: arguments
"TFarRef compatability"

	^ self syncSend: selector withArguments: arguments! !

!FarRef methodsFor: 'errors' stamp: 'ar 1/13/2005 22:16'!
halt
	"This is the typical message to use for inserting breakpoints during 
	debugging. It behaves like halt:, but does not call on halt: in order to 
	avoid putting this message on the stack. Halt is especially useful when 
	the breakpoint message is an arbitrary one."

	Halt signal! !

!FarRef methodsFor: 'errors' stamp: 'ar 1/13/2005 22:16'!
halt: aString 
	"This is the typical message to use for inserting breakpoints during 
	debugging. It creates and schedules a Notifier with the argument, 
	aString, as the label."
	
	Halt new signal: aString! !

!FarRef methodsFor: 'debugging' stamp: 'ar 1/17/2005 23:01'!
inspect
	"Create and schedule an Inspector in which the user can examine the receiver's variables."
	FarRefInspector openOn: self withEvalPane: true! !

!FarRef methodsFor: 'debugging' stamp: 'ar 1/15/2005 13:46'!
instVarAt: index 
	"Primitive. Answer a fixed variable in an object. The numbering of the 
	variables corresponds to the named instance variables. Fail if the index 
	is not an Integer or is not the index of a fixed variable. Essential. See 
	Object documentation whatIsAPrimitive."
	^self primitiveInstVarAt: index! !

!FarRef methodsFor: 'debugging' stamp: 'ar 1/15/2005 13:46'!
instVarAt: anInteger put: anObject 
	"Primitive. Store a value into a fixed variable in the receiver. The 
	numbering of the variables corresponds to the named instance variables. 
	Fail if the index is not an Integer or is not the index of a fixed variable. 
	Answer the value stored as the result. Using this message violates the 
	principle that each object has sovereign control over the storing of 
	values into its instance variables. Essential. See Object documentation 
	whatIsAPrimitive."
	^self primitiveInstVarAt: anInteger put: anObject! !

!FarRef methodsFor: 'accessing' stamp: 'ar 1/15/2005 22:47'!
isFarRef
	^true! !

!FarRef methodsFor: 'debugging' stamp: 'ar 1/13/2005 22:16'!
isKindOf: aClass 
	"Answer whether the class, aClass, is a superclass or class of the receiver."
	self proxyDebuggerMessages ifTrue:[^self send: #isKindOf: with: aClass].
	self class == aClass
		ifTrue: [^true]
		ifFalse: [^self class inheritsFrom: aClass].! !

!FarRef methodsFor: 'debugging' stamp: 'ar 1/13/2005 22:16'!
isMemberOf: aClass 
	"Answer whether the receiver is an instance of the class, aClass."
	self proxyDebuggerMessages ifFalse:[^self class == aClass].
	^self send: #isMemberOf: with: aClass.
! !

!FarRef methodsFor: 'private' stamp: 'ar 1/17/2005 16:36'!
isMorph
	"Too often used and abused. Implement it here to say shut up."
	^false! !

!FarRef methodsFor: 'compatability' stamp: 'mtf 11/2/2011 12:14'!
isResolved
"TFarRef compatability"

	^ true! !

!FarRef methodsFor: 'debugging' stamp: 'ar 1/13/2005 22:16'!
isText
	self proxyDebuggerMessages ifFalse:[^false].
	^self send: #isText! !

!FarRef methodsFor: 'accessing' stamp: 'ar 1/16/2005 16:14'!
island
	"Answer the receiver's island"
	^(myIsland asFarRef: myIsland) valueOn: Processor activeIsland! !

!FarRef methodsFor: 'tracing' stamp: 'mtf 6/1/2010 12:00'!
outboundPointersDo: aBlock
"do aBlock for every object I point to, exactly how the garbage collector would. Adapted from PointerFinder >> #followObject:"

	aBlock value: FarRef.
	aBlock value: myIsland.
	aBlock value: myValue.
! !

!FarRef methodsFor: 'debugging' stamp: 'ar 1/13/2005 22:16'!
perform: selector withArguments: argArray inSuperclass: lookupClass
	"NOTE:  This is just like perform:withArguments:, except that
	the message lookup process begins, not with the receivers's class,
	but with the supplied superclass instead.  It will fail if lookupClass
	cannot be found among the receiver's superclasses.
	Primitive. Essential. See Object documentation whatIsAPrimitive."

	<primitive: 100>
	(selector isMemberOf: Symbol)
		ifFalse: [^ self error: 'selector argument must be a Symbol'].
	(selector numArgs = argArray size)
		ifFalse: [^ self error: 'incorrect number of arguments'].
	(self class == lookupClass or: [self class inheritsFrom: lookupClass])
		ifFalse: [^ self error: 'lookupClass is not in my inheritance chain'].
	self primitiveFailed! !

!FarRef methodsFor: 'primitives' stamp: 'ar 1/13/2005 22:16'!
primitiveBasicAt: index 
	"Primitive. Assumes receiver is indexable. Answer the value of an 
	indexable element in the receiver. Fail if the argument index is not an 
	Integer or is out of bounds. Essential. Do not override in a subclass. See 
	Object documentation whatIsAPrimitive."

	<primitive: 60>
	index isInteger ifTrue: [self error: 'subscript is out of bounds: ' , index printString].
	index isNumber
		ifTrue: [^self primitiveBasicAt: index asInteger]
		ifFalse: [self error: 'only integers should be used as indices']! !

!FarRef methodsFor: 'primitives' stamp: 'ar 1/13/2005 22:16'!
primitiveBasicAt: index put: value
	"Primitive. Assumes receiver is indexable. Store the second argument 
	value in the indexable element of the receiver indicated by index. Fail 
	if the index is not an Integer or is out of bounds. Or fail if the value is 
	not of the right type for this kind of collection. Answer the value that 
	was stored. Essential. Do not override in a subclass. See Object 
	documentation whatIsAPrimitive."

	<primitive: 61>
	index isInteger
		ifTrue: [(index >= 1 and: [index <= self size])
					ifTrue: [self error: 'Improper store into indexable object']
					ifFalse: [self error: 'subscript is out of bounds: ' , index printString]].
	index isNumber
		ifTrue: [^self primitiveBasicAt: index asInteger put: value]
		ifFalse: [self error: 'only integers should be used as indices']! !

!FarRef methodsFor: 'primitives' stamp: 'ar 1/13/2005 22:16'!
primitiveClone

	<primitive: 148>
	self primitiveFailed! !

!FarRef methodsFor: 'primitives' stamp: 'ar 1/13/2005 22:16'!
primitiveInstVarAt: index 
	<primitive: 73>
	^self primitiveBasicAt: index - self class instSize! !

!FarRef methodsFor: 'primitives' stamp: 'ar 1/13/2005 22:16'!
primitiveInstVarAt: anInteger put: anObject 
	<primitive: 74>
	^self primitiveBasicAt: anInteger - self class instSize put: anObject! !

!FarRef methodsFor: 'debugging' stamp: 'mtf 11/2/2011 13:18'!
printOn: aStream
	"Append to the argument, aStream, a sequence of characters that  
	identifies the receiver."

	| title |
	self proxyDebuggerMessages ifTrue:[
		aStream nextPutAll: 'Far:['.
		myValue printOn: aStream. "aStream could be converted to a TFarRef if sent, so don't"
		aStream nextPutAll: ']'.
		^self].
	title := self class name.
	aStream
		nextPutAll: (title first isVowel ifTrue: ['an '] ifFalse: ['a ']);
		nextPutAll: title! !

!FarRef methodsFor: 'debugging' stamp: 'tk 6/6/2005 14:15'!
printString
	"Answer a String whose characters are a description of the receiver. 
	If you want to print without a character limit, use fullPrintString."
	self proxyDebuggerMessages ifTrue:[
		^'Far ', myIsland name asString, ':[', (self send: #printString),']'].
	^ self printStringLimitedTo: 50000! !

!FarRef methodsFor: 'debugging' stamp: 'ar 1/13/2005 22:16'!
printStringLimitedTo: limit
	"Answer a String whose characters are a description of the receiver.
	If you want to print without a character limit, use fullPrintString."
	| limitedString |
	self proxyDebuggerMessages ifTrue:[^self send: #printStringLimitedTo: with: limit].
	limitedString := String streamContents: [:s | self printOn: s] limitedTo: limit.
	limitedString size < limit ifTrue: [^ limitedString].
	^ limitedString , '...etc...'! !

!FarRef methodsFor: 'private' stamp: 'ar 1/17/2005 16:37'!
privateIsland
	"Here exclusively for the purpose of being able to uninstall islands"
	^myIsland! !

!FarRef methodsFor: 'private' stamp: 'ar 1/14/2005 22:35'!
privateValue: anObject island: anIsland
	myValue := anObject.
	myIsland := anIsland.! !

!FarRef methodsFor: 'private' stamp: 'ar 1/16/2005 16:29'!
proxyDebuggerMessages
	"Answer whether we should proxy messages sent via the debugger and/or inspector. Generally this should be the case as ANY messages should be passed to the island but that makes it very hard to debug the messaging level. This hook allows us to either disable all of the messages sent during debugging or alternatively, just those for specific objects."
	^true! !

!FarRef methodsFor: 'messaging' stamp: 'mtf 11/2/2011 12:07'!
send: selector

	"TFarRef compatability"
	selector isBlock ifTrue: [^ selector value: self].

	^self send: selector withArguments: #()! !

!FarRef methodsFor: 'messaging' stamp: 'ar 1/14/2005 22:37'!
send: selector with: arg1
	^self send: selector withArguments: (Array with: arg1)! !

!FarRef methodsFor: 'messaging' stamp: 'ar 1/14/2005 22:37'!
send: selector with: arg1 with: arg2
	^self send: selector withArguments: (Array with: arg1 with: arg2)! !

!FarRef methodsFor: 'messaging' stamp: 'ar 1/14/2005 22:38'!
send: selector withArguments: args
	"Send the selector, aSymbol, to the receiver with arguments in argArray."
	^self syncSend: selector withArguments: args! !

!FarRef methodsFor: 'private' stamp: 'ar 6/16/2005 19:59'!
startUpInIslandReader: reader
	"Do nothing"! !

!FarRef methodsFor: 'messaging' stamp: 'ar 1/17/2005 22:57'!
syncReturn: result from: dstIsland to: srcIsland
	^(IslandArgumentCopier new from: dstIsland to: srcIsland) pass: result.! !

!FarRef methodsFor: 'messaging' stamp: 'sn 2/22/2018 12:11'!
syncSend: selector withArguments: arguments 
	"Send the selector, aSymbol, to the receiver with arguments in arguments."
	| copier args dstIsland srcIsland result arg |
	srcIsland := Processor activeIsland.
	dstIsland := myIsland.
	srcIsland == dstIsland 
		ifTrue:[^myValue perform: selector withArguments: arguments].

	copier := nil.
	args := arguments shallowCopy.
	1 to: arguments size do:[:i|
		arg := arguments at: i.
		(IslandArgumentCopier isPassByIdentity: arg) ifFalse:[
			copier ifNil:[copier := IslandArgumentCopier new from:  srcIsland to: dstIsland].
			arg := copier pass: arg.
		].
		args at: i put: arg.
	].
	result := Processor activeProcess activateIslandWithEscape: myIsland
		during: [myValue perform: selector withArguments: args].
	^self syncReturn: result from: dstIsland to: srcIsland! !

!FarRef methodsFor: 'accessing' stamp: 'ar 1/14/2005 22:29'!
valueOn: anIsland
	"Answer the object/value I would have on the given island"
	anIsland == myIsland ifTrue:[^myValue] ifFalse:[^self].! !

!FarRef methodsFor: 'accessing' stamp: 'ar 1/17/2005 21:19'!
varValueOn: anIsland
	"Answer the object/value I would have on the given island.
	This version is specific for variables handling copy on read."
	anIsland == myIsland ifTrue:[^myValue].
	"Do a quick check for #passByProxy: ... "
	myValue class howToPassAsArgument == #passByProxy: ifTrue:[^self].
	" ... but otherwise use slow, generic mechanism."
	anIsland isFarRef ifTrue:[^self error: 'Cannot compute value on far island'].
	^self syncReturn: myValue from: myIsland to: anIsland! !

!FarRef methodsFor: 'copying' stamp: 'tk 6/2/2005 10:22'!
veryDeepCopyWith: aDeepCopier
	"I refer to an object in another project.  Replicate that object here, with no references to the source project."

	| new |
	aDeepCopier references at: self ifPresent: [:newer | ^ newer]. 	"already did him"
	aDeepCopier forbiddenIsland = myIsland
		ifTrue: [new := (IslandDeepCopier new from: myIsland to: Processor activeIsland) 
					pass: myValue.	"always copies"
				aDeepCopier references at: self put: new.	"remember"
				aDeepCopier newGuys add: new.	"remember"
				^ new]
		ifFalse: [^ self]	"always OK to share FarRefs"! !

!BitBlt class methodsFor: '*Islands' stamp: 'ar 1/16/2005 21:55'!
howToPassAsArgument
	"passByIdentity: for rendering purposes"
	^#passByIdentity:! !

!ProcessorScheduler class methodsFor: '*Islands' stamp: 'ar 1/16/2005 21:59'!
howToPassAsArgument
	"passByIdentity: or else none of e.g., 'Processor activeMumble' will work"
	^#passByIdentity:! !

!ProcessorScheduler methodsFor: '*Islands' stamp: 'mtf 1/18/2011 20:33'!
activeIsland
	"Answer the currently active Island."
	^ self activeProcess island ifNil:[Island default]! !

!ProcessorScheduler methodsFor: '*Islands' stamp: 'mtf 1/18/2011 20:33'!
activeIsland: anIsland
	"Answer the currently active Island."
	self activeProcess island: anIsland! !

!WeakRegistry class methodsFor: '*Islands' stamp: 'ar 1/16/2005 22:03'!
howToPassAsArgument
	"passByIdentity: since WeakRegistries are shared resources"
	^#passByIdentity:! !

!Number class methodsFor: '*Islands' stamp: 'ar 1/16/2005 21:58'!
howToPassAsArgument
	"Numbers are immutable"
	^#passByIdentity:! !

!MorphicEvent class methodsFor: '*Islands' stamp: 'ar 1/17/2005 21:18'!
howToPassAsArgument
	"Really, this ought to be passByCopy: but some events use handler returns etc."
	^#passByProxy:! !

!Class methodsFor: '*Islands' stamp: 'ar 6/16/2005 19:40'!
classVersionID
	"Answer the receiver's version ID"
	^nil! !

!Class methodsFor: '*Islands' stamp: 'ar 6/16/2005 19:40'!
exportOnIslandWriter: writer
	^writer exportClass: self! !

!ZipFileMember methodsFor: '*Islands' stamp: 'ar 6/16/2005 17:55'!
byteStream
	"Answer a suitable byte stream for the contents of the receiver.
	Intended to avoid *any* temporary decompression and conversion."
	| decoder |
	self rewindData.
	(compressionMethod = CompressionStored) ifTrue:[
		"Return uncompressed stream"
		^stream
	].
	(compressionMethod = CompressionDeflated) ifTrue:[
		"Return compressed stream"
		decoder := ZipReadStream on: stream.
		decoder expectedCrc: self crc32.
		^decoder
	].
	^self error: 'Unknown compression method'! !

!Character class methodsFor: '*Islands' stamp: 'ar 1/16/2005 21:56'!
howToPassAsArgument
	"passByIdentity: since characters are immutable"
	^#passByIdentity:! !

!Character methodsFor: '*Islands' stamp: 'ar 6/16/2005 19:40'!
exportOnIslandWriter: writer
	^writer exportCharacter: self! !

!UndefinedObject class methodsFor: '*Islands' stamp: 'ar 1/16/2005 22:03'!
howToPassAsArgument
	"passByIdentity: since nil is immutable"
	^#passByIdentity:! !

!UndefinedObject methodsFor: '*Islands' stamp: 'ar 1/17/2005 21:23'!
asFarRef: obj
	"Fallback position"
	^obj! !

!UndefinedObject methodsFor: '*Islands' stamp: 'ar 5/22/2005 15:34'!
classOf: obj
	"Fallback position"
	^obj class! !

!UndefinedObject methodsFor: '*Islands' stamp: 'ar 6/16/2005 19:45'!
exportOnIslandWriter: writer
	^writer exportNil: self! !

!UndefinedObject methodsFor: '*Islands' stamp: 'ar 1/16/2005 09:58'!
new: aClass
	"Fallback position"
	^aClass new! !

!Canvas class methodsFor: '*Islands' stamp: 'mtf 1/24/2011 17:48'!
howToPassAsArgument
	^ #passByIdentity:! !

!Integer methodsFor: '*Islands' stamp: 'ar 6/16/2005 19:42'!
exportOnIslandWriter: writer
	^writer exportInteger: self! !

Error subclass: #CheckpointFailure
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Islands-Files'!
Island initialize!
IslandReader initialize!
IslandWriter initialize!
